#include <Pilot.h>
#include <UI/ScrDriver.h>
#include "2bit.h"
#include "globals.h"
#include "callback.h"

#ifndef NOTWOBIT

#define FTR_2BIT 25  /* two bits :-) */

#define MAGICVAL 0x2b828184

static volatile DWord magic = MAGICVAL;

typedef struct {
    const void *SSA;
    unsigned char VPW;
    unsigned char PICF;
    unsigned char LBAR;
    unsigned char FRCM;
    Word GPMR;
    unsigned char POSR;
} DispParams;

static WinHandle twobitwindow = NULL;
static DispParams savedispparams;

typedef void (*ScrLineRoutineType)(WinPtr, SWord, SWord, SWord, SWord);
static ScrLineRoutineType ScrLineRoutine_1bit = NULL;

typedef void (*ScrRectangleRoutineType)(WinPtr, SWord, SWord, SWord, SWord);
static ScrRectangleRoutineType ScrRectangleRoutine_1bit = NULL;

typedef void (*ScrDrawCharsType)(WinPtr, SWord, SWord, SWord, SWord, SWord,
    SWord, SWord, SWord, CharPtr, Word, FontPtr);
static ScrDrawCharsType ScrDrawChars_1bit = NULL;

typedef Err (*ScrCopyRectangleType)(WinPtr, WinPtr, SWord, SWord, SWord,
    SWord, SWord, SWord);
static ScrCopyRectangleType ScrCopyRectangle_1bit = NULL;

static int in_2bit = 0;

static const unsigned short grlmask[8] = {
    0xffff, 0x3fff, 0x0fff, 0x03ff, 0x00ff, 0x003f, 0x000f, 0x0003 };
static const unsigned short grrmask[8] = {
    0xc000, 0xf000, 0xfc00, 0xff00, 0xffc0, 0xfff0, 0xfffc, 0xffff };
static const unsigned short bw2gr[256] = {
    0x0000, 0x0003, 0x000c, 0x000f, 0x0030, 0x0033, 0x003c, 0x003f,
    0x00c0, 0x00c3, 0x00cc, 0x00cf, 0x00f0, 0x00f3, 0x00fc, 0x00ff,
    0x0300, 0x0303, 0x030c, 0x030f, 0x0330, 0x0333, 0x033c, 0x033f,
    0x03c0, 0x03c3, 0x03cc, 0x03cf, 0x03f0, 0x03f3, 0x03fc, 0x03ff,
    0x0c00, 0x0c03, 0x0c0c, 0x0c0f, 0x0c30, 0x0c33, 0x0c3c, 0x0c3f,
    0x0cc0, 0x0cc3, 0x0ccc, 0x0ccf, 0x0cf0, 0x0cf3, 0x0cfc, 0x0cff,
    0x0f00, 0x0f03, 0x0f0c, 0x0f0f, 0x0f30, 0x0f33, 0x0f3c, 0x0f3f,
    0x0fc0, 0x0fc3, 0x0fcc, 0x0fcf, 0x0ff0, 0x0ff3, 0x0ffc, 0x0fff,
    0x3000, 0x3003, 0x300c, 0x300f, 0x3030, 0x3033, 0x303c, 0x303f,
    0x30c0, 0x30c3, 0x30cc, 0x30cf, 0x30f0, 0x30f3, 0x30fc, 0x30ff,
    0x3300, 0x3303, 0x330c, 0x330f, 0x3330, 0x3333, 0x333c, 0x333f,
    0x33c0, 0x33c3, 0x33cc, 0x33cf, 0x33f0, 0x33f3, 0x33fc, 0x33ff,
    0x3c00, 0x3c03, 0x3c0c, 0x3c0f, 0x3c30, 0x3c33, 0x3c3c, 0x3c3f,
    0x3cc0, 0x3cc3, 0x3ccc, 0x3ccf, 0x3cf0, 0x3cf3, 0x3cfc, 0x3cff,
    0x3f00, 0x3f03, 0x3f0c, 0x3f0f, 0x3f30, 0x3f33, 0x3f3c, 0x3f3f,
    0x3fc0, 0x3fc3, 0x3fcc, 0x3fcf, 0x3ff0, 0x3ff3, 0x3ffc, 0x3fff,
    0xc000, 0xc003, 0xc00c, 0xc00f, 0xc030, 0xc033, 0xc03c, 0xc03f,
    0xc0c0, 0xc0c3, 0xc0cc, 0xc0cf, 0xc0f0, 0xc0f3, 0xc0fc, 0xc0ff,
    0xc300, 0xc303, 0xc30c, 0xc30f, 0xc330, 0xc333, 0xc33c, 0xc33f,
    0xc3c0, 0xc3c3, 0xc3cc, 0xc3cf, 0xc3f0, 0xc3f3, 0xc3fc, 0xc3ff,
    0xcc00, 0xcc03, 0xcc0c, 0xcc0f, 0xcc30, 0xcc33, 0xcc3c, 0xcc3f,
    0xccc0, 0xccc3, 0xcccc, 0xcccf, 0xccf0, 0xccf3, 0xccfc, 0xccff,
    0xcf00, 0xcf03, 0xcf0c, 0xcf0f, 0xcf30, 0xcf33, 0xcf3c, 0xcf3f,
    0xcfc0, 0xcfc3, 0xcfcc, 0xcfcf, 0xcff0, 0xcff3, 0xcffc, 0xcfff,
    0xf000, 0xf003, 0xf00c, 0xf00f, 0xf030, 0xf033, 0xf03c, 0xf03f,
    0xf0c0, 0xf0c3, 0xf0cc, 0xf0cf, 0xf0f0, 0xf0f3, 0xf0fc, 0xf0ff,
    0xf300, 0xf303, 0xf30c, 0xf30f, 0xf330, 0xf333, 0xf33c, 0xf33f,
    0xf3c0, 0xf3c3, 0xf3cc, 0xf3cf, 0xf3f0, 0xf3f3, 0xf3fc, 0xf3ff,
    0xfc00, 0xfc03, 0xfc0c, 0xfc0f, 0xfc30, 0xfc33, 0xfc3c, 0xfc3f,
    0xfcc0, 0xfcc3, 0xfccc, 0xfccf, 0xfcf0, 0xfcf3, 0xfcfc, 0xfcff,
    0xff00, 0xff03, 0xff0c, 0xff0f, 0xff30, 0xff33, 0xff3c, 0xff3f,
    0xffc0, 0xffc3, 0xffcc, 0xffcf, 0xfff0, 0xfff3, 0xfffc, 0xffff
};

static void update_2bit(unsigned char *base, SWord x1, SWord y1,
    SWord x2, SWord yextent)
{
    unsigned char *bwstartline = base + y1*20;
    unsigned short *grstartline = ((unsigned short *)twobitwindow->displayAddr)
				    + y1*20;
    unsigned short lmask = grlmask[x1 & 7];
    unsigned short rmask = grrmask[x2 & 7];

    if (in_2bit) return;
    for(;yextent;--yextent) {
	if ((x1 >> 3) == (x2 >> 3)) {
	    /* Only one byte */
	    unsigned short mask = (lmask & rmask);
	    unsigned short grval = grstartline[x1 >> 3] & ~mask;
	    grval |= (bw2gr[bwstartline[x1>>3]] & mask);
	    grstartline[x1 >> 3] = grval;
	} else {
	    SWord i;
	    /* The first (partial) byte */
	    unsigned short grval = grstartline[x1 >> 3] & ~lmask;
	    grval |= (bw2gr[bwstartline[x1>>3]] & lmask);
	    grstartline[x1 >> 3] = grval;
	    /* Middle bytes */
	    for(i=(x1>>3)+1;i<(x2>>3);++i) {
		grstartline[i] = bw2gr[bwstartline[i]];
	    }
	    /* The last (partial) byte */
	    grval = grstartline[x2 >> 3] & ~rmask;
	    grval |= (bw2gr[bwstartline[x2>>3]] & rmask);
	    grstartline[x2 >> 3] = grval;
	}
	bwstartline += 20;
	grstartline += 20;
    }
}

static void ScrLineRoutine_2bit(WinPtr pWindow, SWord x1, SWord y1, SWord x2,
    SWord y2)
{
    CALLBACK_INTERAPP_PROLOGUE(APPTYPE, FTR_2BIT, magic, MAGICVAL)

    ScrLineRoutine_1bit(pWindow, x1, y1, x2, y2);
    if (x1 > x2) {
	SWord t = x1;
	x1 = x2;
	x2 = t;
    }
    if (y1 > y2) {
	SWord t = y1;
	y1 = y2;
	y2 = t;
    }
    if (! pWindow->windowFlags.offscreen) {
	update_2bit(pWindow->displayAddr,x1,y1,x2,y2-y1+1);
    }

    CALLBACK_INTERAPP_EPILOGUE
}

static void ScrRectangleRoutine_2bit(WinPtr pWindow, SWord x, SWord y,
    SWord extentX, SWord extentY)
{
    CALLBACK_INTERAPP_PROLOGUE(APPTYPE, FTR_2BIT, magic, MAGICVAL)

    ScrRectangleRoutine_1bit(pWindow, x, y, extentX, extentY);
    if (! pWindow->windowFlags.offscreen) {
	update_2bit(pWindow->displayAddr,x,y,x+extentX-1,extentY);
    }

    CALLBACK_INTERAPP_EPILOGUE
}

static void ScrDrawChars_2bit(WinPtr pWindow, SWord xLoc, SWord yLoc,
    SWord xExtent, SWord yExtent, SWord clipTop, SWord clipLeft,
    SWord clipBottom, SWord clipRight, CharPtr chars, Word len,
    FontPtr fontPtr)
{
    CALLBACK_INTERAPP_PROLOGUE(APPTYPE, FTR_2BIT, magic, MAGICVAL)

    ScrDrawChars_1bit(pWindow, xLoc, yLoc, xExtent, yExtent, clipTop,
	clipLeft, clipBottom, clipRight, chars, len, fontPtr);
    if (! pWindow->windowFlags.offscreen) {
	update_2bit(pWindow->displayAddr,xLoc,yLoc,xLoc+xExtent-1,yExtent);
    }

    CALLBACK_INTERAPP_EPILOGUE
}

static Err ScrCopyRectangle_2bit(WinPtr sourceWindow, WinPtr destWindow,
    SWord fromX, SWord fromY, SWord toX, SWord toY, SWord bitCount,
    SWord lineCount) 
{
    Err retval;
    CALLBACK_INTERAPP_PROLOGUE(APPTYPE, FTR_2BIT, magic, MAGICVAL)

    retval = ScrCopyRectangle_1bit(sourceWindow, destWindow, fromX, fromY,
	toX, toY, bitCount, lineCount);
    if (! destWindow->windowFlags.offscreen) {
	update_2bit(destWindow->displayAddr,toX,toY,
	    toX+bitCount-1,lineCount);
    }

    CALLBACK_INTERAPP_EPILOGUE
    return retval;
}

/* Set up the display in either greyscale or B/W mode.  vpw is the virtual
   page width _in words_ (word = 16 bits).  The info used here was derived
   from code in a _PDA Developers_ article by Edward Keyes. */
static void setgrey(DispParams *curparams, const void *newSSA,
    DispParams *newparams)
{
    const void * volatile *SSA = (const void**)0xfffffa00;
    volatile unsigned char *CKCON = (unsigned char *)0xfffffa27;
    volatile unsigned char *VPW = (unsigned char *)0xfffffa05;
    volatile unsigned char *PICF = (unsigned char *)0xfffffa20;
    volatile unsigned char *LBAR = (unsigned char *)0xfffffa29;
    volatile unsigned char *FRCM = (unsigned char *)0xfffffa31;
    volatile unsigned char *POSR = (unsigned char *)0xfffffa2d;
    volatile Word *GPMR = (Word *)0xfffffa32;
    unsigned gray0 = 0, gray1 = 3, gray2 = 4, gray3 = 6;

    /* Save the old display parameters */
    if (curparams) {
	curparams->SSA = *SSA;
	curparams->VPW = *VPW;
	curparams->PICF = *PICF;
	curparams->LBAR = *LBAR;
	curparams->FRCM = *FRCM;
	curparams->GPMR = *GPMR;
	curparams->POSR = *POSR;
    }

    if (newparams && !newparams->SSA && !newSSA) return;

    /* Turn off the display while we fiddle */
    *CKCON &= 0x7f;
    SysTaskDelay(2);

    if (newSSA) {
	*SSA = newSSA;
    }

    if (newparams) {
	/* We are restoring parameters */
	if (!newSSA) {
	    *SSA = newparams->SSA;
	}
	*VPW = newparams->VPW;
	*PICF = newparams->PICF;
	*LBAR = newparams->LBAR;
	*FRCM = newparams->FRCM;
	*GPMR = newparams->GPMR;
	*POSR = newparams->POSR;
    } else {
	/* We are going to greyscale */
        *VPW = 20;
        *PICF |= 0x01;
        *LBAR = 22;
        *FRCM = 0xb9;
        *GPMR = gray2 + (gray3<<4) + (gray0<<8) + (gray1<<12);
        *POSR &= 0xf0;
    }
    /*
	If we ever have to forcefully go to b/w mode, do this:

        *VPW = 10;
        *PICF &= 0xfe;
        *LBAR = 10;
        *POSR &= 0xf0;
    */

    /* Turn the display back on */
    SysTaskDelay(2);
    *CKCON |= 0x80;
}

void WinDraw2Bitmap(BitmapPtr image, SWord x, SWord y)
{
    WinHandle olddraw;
    RectangleType oldclip1bit, oldclip2bit, newclip;
    int save_2bit = in_2bit;
    
    in_2bit = 1;
    olddraw = WinGetDrawWindow();
    WinGetClip(&oldclip1bit);
    newclip.topLeft.x = 2*oldclip1bit.topLeft.x;
    newclip.topLeft.y = oldclip1bit.topLeft.y;
    newclip.extent.x = 2*oldclip1bit.extent.x;
    newclip.extent.y = oldclip1bit.extent.y;
    WinSetDrawWindow(twobitwindow);
    WinGetClip(&oldclip2bit);
    WinSetClip(&newclip);
    WinDrawBitmap(image, 2*x, y);
    WinSetClip(&oldclip2bit);
    WinSetDrawWindow(olddraw);
    in_2bit = save_2bit;
}

void WinCopy2BitRectangle(WinHandle win, RectanglePtr srcr, SWord dx, SWord dy,
    ScrOperation mode)
{
    RectangleType srcr2 = *srcr;
    int save_2bit = in_2bit;

    srcr2.topLeft.x *= 2;
    srcr2.extent.x *= 2;
    in_2bit = 1;
    WinCopyRectangle(win, win, srcr, dx, dy, mode);
    WinCopyRectangle(twobitwindow, twobitwindow, &srcr2, dx*2, dy, mode);
    in_2bit = save_2bit;
}

void enable_2bit(void)
{
    Err err;
    WinHandle olddraw;

    /* Allocate a window */
    twobitwindow = WinCreateOffscreenWindow(320,160,screenFormat,&err);
    olddraw = WinGetDrawWindow();
    WinSetDrawWindow(twobitwindow);
    WinEraseWindow();
    WinSetDrawWindow(olddraw);

    /* Set up trap patches */
    ScrLineRoutine_1bit =
	(ScrLineRoutineType)GSysDispatchTableP[sysTrapScrLineRoutine-sysTrapBase];
    GSysDispatchTableP[sysTrapScrLineRoutine-sysTrapBase] =
	(Ptr)ScrLineRoutine_2bit;

    ScrRectangleRoutine_1bit =
	(ScrRectangleRoutineType)GSysDispatchTableP[sysTrapScrRectangleRoutine-sysTrapBase];
    GSysDispatchTableP[sysTrapScrRectangleRoutine-sysTrapBase] =
	(Ptr)ScrRectangleRoutine_2bit;

    ScrDrawChars_1bit =
	(ScrDrawCharsType)GSysDispatchTableP[sysTrapScrDrawChars-sysTrapBase];
    GSysDispatchTableP[sysTrapScrDrawChars-sysTrapBase] =
	(Ptr)ScrDrawChars_2bit;

    ScrCopyRectangle_1bit =
	(ScrCopyRectangleType)GSysDispatchTableP[sysTrapScrCopyRectangle-sysTrapBase];
    GSysDispatchTableP[sysTrapScrCopyRectangle-sysTrapBase] =
	(Ptr)ScrCopyRectangle_2bit;

    /* Go 2-bit */
    setgrey(&savedispparams, (const void *)twobitwindow->displayAddr, NULL);
    REGISTER_INTERAPP_CALLBACK(APPTYPE, FTR_2BIT)
}

void disable_2bit(void)
{
    /* Go 1-bit */
    DEREGISTER_INTERAPP_CALLBACK(APPTYPE, FTR_2BIT)
    setgrey(NULL, NULL, &savedispparams);

    /* Restore traps */
    GSysDispatchTableP[sysTrapScrLineRoutine-sysTrapBase] =
	(Ptr)ScrLineRoutine_1bit;

    GSysDispatchTableP[sysTrapScrRectangleRoutine-sysTrapBase] =
	(Ptr)ScrRectangleRoutine_1bit;

    GSysDispatchTableP[sysTrapScrDrawChars-sysTrapBase] =
	(Ptr)ScrDrawChars_1bit;

    GSysDispatchTableP[sysTrapScrCopyRectangle-sysTrapBase] =
	(Ptr)ScrCopyRectangle_1bit;

    /* Delete the window */
    WinDeleteWindow(twobitwindow, false);
}

#endif /* !NOTWOBIT */
