#include <Pilot.h>
#include "globals.h"
#include "callback.h"
#include "send.h"
#include "aggform.h"
#include "editmenu.h"
#include "utils.h"

static VoidHand hAggPopupEntries = NULL;
/* handle to the PopupEntries. Init'ed out here multiple calls of the
   aggform event loop use it, specifically open and close */

static Word *aggDataEntries = NULL;

static int enclen(CharPtr u)
{
    int len = 0;
    if (!u) return 0;
    while(*u) {
	if (*u < '!' || *u > 'z' || *u == '%' || *u == '&' || *u == '?' ||
	    *u == '=' || *u == '/' || *u == ':' || *u == ';' || *u == '*' ||
	    *u == '^' || *u == '+' || *u == '#') {
	    len += 2;
	}
	++len;
	++u;
    }
    return len;
}

static void enccopy(CharPtr d, CharPtr _s)
{
    static const char * const h = "0123456789abcdef";
    unsigned char *s = _s;
    if (s) {
	while(*s) {
	    if (*s < '!' || *s > 'z' || *s == '%' || *s == '&' || *s == '?' ||
		*s == '=' || *s == '/' || *s == ':' || *s == ';' || *s == '*'
		|| *s == '^' || *s == '+' || *s == '#') {
		*(d++) = '%';
		*(d++) = h[(*s)>>4];
		*(d++) = h[(*s)&0x0f];
	    } else {
		*(d++) = *s;
	    }
	    ++s;
	}
    }
    *d = '\0';
}

Boolean AggFormHandleEvent (EventPtr e) {
  Boolean handled = false;
  FormPtr frm;
  FieldPtr AggNameField, AggValField;
  CharPtr aggname, aggval;
  int aggnamelen, aggvallen;
  VoidHand hAgg;
  CharPtr AggPtr;
  SWord aggDataSize, aggReadSize;

  CALLBACK_PROLOGUE

  switch (e->eType) {
  case frmOpenEvent:
    frm = FrmGetActiveForm();
    {
      ListPtr pAggPopList;

      /* aggPref involves the stored (actual) preference in database*/
      /* aggData involves the retreived (copy of) preference */
      aggDataSize = 0; 
      aggReadSize = PrefGetAppPreferences(APPTYPE, AGG_LIST_PREF,
					       NULL, &aggDataSize, true);
      /* aggDataSize is the actual size of the pref; aggReadSize is
         the read amount */

      pAggPopList = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
							      AggSelectPopList));

      if ((aggReadSize != noPreferenceFound) && (aggDataSize > 0)) {
	UInt aggDataNumEntries;
	char **aggPopupEntries;

	/* getting the data into a usable form */
	aggDataEntries = MemPtrNew((ULong) aggDataSize);
	ErrFatalDisplayIf(!aggDataEntries, "Cannot allocate memory");
	(void) PrefGetAppPreferences(APPTYPE, AGG_LIST_PREF,
				   aggDataEntries, &aggDataSize, true);
	aggDataNumEntries = aggDataEntries[0];
	hAggPopupEntries = SysFormPointerArrayToStrings((char *)
							(aggDataEntries + 1),
							aggDataNumEntries);
							

	/* -> at this point agg PopupEntries is a unlocked handle */
	aggPopupEntries = MemHandleLock(hAggPopupEntries);
	
	LstSetListChoices(pAggPopList, aggPopupEntries, aggDataNumEntries);
	if (aggDataNumEntries < 5) {
	  LstSetHeight(pAggPopList, aggDataNumEntries);
	} else {
	  LstSetHeight(pAggPopList, 5);
	}

	MemHandleLock(hAggPopupEntries);
      } else {
	/* NULL */
      }
    }

    /* Possibly restore the saved name and value from last time */
    aggDataSize = 0;
    aggReadSize = PrefGetAppPreferences(APPTYPE, AGG_SAVE_PREF, NULL,
					&aggDataSize, true);
    if ((aggReadSize != noPreferenceFound) && (aggDataSize > 0)) {
	/* Get the saved values */
	ULong nlen, vlen;
	FieldPtr pAggNameFld =
	  FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AggNameFld));
	FieldPtr pAggValFld =
	  FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AggValFld));
	CharPtr saveddata = MemPtrNew((ULong) aggDataSize);
	ErrFatalDisplayIf(!saveddata, "Cannot allocation memory");
	(void) PrefGetAppPreferences(APPTYPE, AGG_SAVE_PREF, saveddata,
				     &aggDataSize, true);
	nlen = StrLen(saveddata);
	SetFld(pAggNameFld, saveddata, nlen);
	vlen = StrLen(saveddata+nlen+1);
	SetFld(pAggValFld, saveddata+nlen+1, vlen);
	MemPtrFree(saveddata);
    }

    FrmDrawForm(frm);
    FrmSetFocus(frm, FrmGetObjectIndex(frm, AggNameFld));

    handled = true;
    break;

  case menuEvent:
    MenuEraseStatus(NULL);
    if (EditMenuHandleEvent(e->data.menu.itemID, FrmGetActiveForm())) {
	handled = true;
    } else {
	switch(e->data.menu.itemID) {
	}
    }
    handled = true;
    break;

  case keyDownEvent:
    switch(e->data.keyDown.chr) {
    Word focus;
    Word fldid;
    case nextFieldChr:
    case prevFieldChr:
	/* Go to the other field */
	frm = FrmGetActiveForm();
	focus = FrmGetFocus(frm);
	if (focus == noFocus) break;
	fldid = FrmGetObjectId(frm, focus);
	if (fldid == AggNameFld && e->data.keyDown.chr == nextFieldChr) {
	    FrmSetFocus(frm, FrmGetObjectIndex(frm, AggValFld));
	} else if (fldid == AggValFld && e->data.keyDown.chr == prevFieldChr) {
	    FrmSetFocus(frm, FrmGetObjectIndex(frm, AggNameFld));
	}
	handled = true;
	break;
    }
    break;

  case ctlSelectEvent:
    switch(e->data.ctlEnter.controlID) {
    case AggSelectPopCtl:
      frm = FrmGetActiveForm();
      {
	ListPtr pAggPopList =
	  FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AggSelectPopList));
	FieldPtr pAggNameFld =
	  FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AggNameFld));
	short aggSelected = LstPopupList(pAggPopList);
	if (aggSelected != -1) {
	  CharPtr aggSelectedText =
	    LstGetSelectionText(pAggPopList, aggSelected);
	  SetFld(pAggNameFld, aggSelectedText, StrLen(aggSelectedText));
	  FldDrawField(pAggNameFld);
	}
      }
    break;

    case AggFormOK:
    case AggFormHelp:
      frm = FrmGetActiveForm();
      AggNameField = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AggNameFld));
      AggValField = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AggValFld));

      /* get the name and the value and copy them into a buffer */
      if (e->data.ctlEnter.controlID == AggFormOK) {
	  aggname = FldGetTextPtr(AggNameField);
	  aggnamelen = enclen(aggname);
	  if (!aggnamelen) {
	    aggname = "help";
	    aggnamelen = 4;
	  }
	  aggval = FldGetTextPtr(AggValField);
	  aggvallen = enclen(aggval);
      } else {
	aggname = "help";
	aggnamelen = 4;
	aggval = NULL;
	aggvallen = 0;
      }
      hAgg = MemHandleNew((ULong) (8 + aggnamelen + aggvallen));
      ErrFatalDisplayIf(!hAgg, "Out of memory!");
      AggPtr = MemHandleLock(hAgg);
      StrCopy(AggPtr, "agg://");
      enccopy(AggPtr+6, aggname);
      AggPtr[6+aggnamelen] = '/';
      enccopy(AggPtr+7+aggnamelen, aggval);
      MemHandleUnlock(hAgg);

      /* Send the URL, but ignore the local cache */
      /*
      if (transURL) {
	MemHandleFree(transURL);
      }
      transURLLength = 8 + aggnamelen + aggvallen;
      transURL = hAgg; 
      */
      
      if (!SendURL(&conID,  hAgg, true, false, true, true)) {
	MemHandleFree(hAgg);
      }

      if (e->data.ctlEnter.controlID == AggFormOK) {
	  /* Save the current values of the name and val fields */
	  ULong nlen, vlen;
	  CharPtr save;
	  aggname = FldGetTextPtr(AggNameField);
	  aggval = FldGetTextPtr(AggValField);
	  nlen = aggname ? StrLen(aggname) : 0;
	  vlen = aggval ? StrLen(aggval) : 0;
	  save = MemPtrNew(nlen+vlen+2);
	  ErrFatalDisplayIf(!save, "Cannot allocate memory");
	  if (nlen) MemMove(save, aggname, nlen);
	  save[nlen] = '\0';
	  if (vlen) MemMove(save+nlen+1, aggval, vlen);
	  save[nlen+vlen+1] = '\0';
	  PrefSetAppPreferences(APPTYPE, AGG_SAVE_PREF, 0, save, nlen+vlen+2,
				true);
	  MemPtrFree(save);
      }

      /* FALLTHROUGH */

    case AggFormCan:

      /* cleaning up agg listing handle */
      if (hAggPopupEntries) { /* why would this handle be invalid? */
	MemHandleFree(hAggPopupEntries);
      }
      if (aggDataEntries) {
	MemPtrFree(aggDataEntries);
      }
      FrmReturnToForm(0); 
      break;
    }
    handled = true;
    break;

  default:
    break;
  }

  CALLBACK_EPILOGUE

  return handled;
}

static int gotagglist = 0;

int NeedAggList(void)
{
    return !gotagglist;
}

void GotAggList(void *agglist, Word size)
{
    PrefSetAppPreferences(APPTYPE, AGG_LIST_PREF, 0, agglist, size, true);
    gotagglist = 1;
}
