#include <Pilot.h>
#include "globals.h"
#include "cache.h"

/*
 * cacheInfo records look like
 * ULong id of actually data record for the cache (not info) data for that
 * page;
 * Word size of url string
 * DWord expr time
 * followed by the url (null terminated string)
 * followed by the name cached page (null terminated string OR nonexistant) 
 *    (for the edit screen) 
 * right now expration date doesn't work quite right
 * if something in the cache has expired it stays in the cache but
 * will not be returned by InCache.  So it IS taking up memory.
 * not that big deal as it will be pushed out of the cache after a while
 */
VoidHand InCache(char *URL, Word URLSize, DmOpenRef infoDB, DmOpenRef DB) {  
  ULong numRecs;
  ULong i;
  UInt cardNo;
  LocalID localID;
  ULong ID;  
  VoidHand rec;
  void * recPtr;
  UInt index;

  DmOpenDatabaseInfo(infoDB, &localID, NULL, NULL, &cardNo, NULL);
  DmDatabaseSize(cardNo, localID, &numRecs, NULL, NULL);
  
  if (numRecs) {    
    for (i = 0; i < numRecs; i++) {
      Boolean match;
      rec = DmGetRecord(infoDB, i);
      recPtr = MemHandleLock(rec);
      match = false;
      if (*((Word *) (recPtr + sizeof(ULong))) == URLSize) {
	ULong t;
	
	t = *((ULong *) (recPtr + sizeof(ULong) + sizeof(Word)));
	if (!t || (t >= TimGetSeconds())) {
	  match = !(MemCmp(URL, 
			   recPtr+sizeof(ULong)+sizeof(Word)+sizeof(ULong), 
			   (ULong) URLSize));
	}
      }
      ID = *((ULong*) recPtr);
      MemHandleUnlock(rec);
      DmReleaseRecord(infoDB, i, false);      
      if (match) {
        Err err;
	DmMoveRecord(infoDB, (UInt) i, 0);
	err = DmFindRecordByID(DB, ID, &index);
	ErrFatalDisplayIf(err, "Cache corrupted!");
	rec = DmQueryRecord(DB, index);
	return rec;
      }
    }
  }
  return 0;
}

/* if url was in the cache, remove it, otherise do nothing */
void RemoveFromCache(char *URL, Word URLSize, DmOpenRef infoDB, 
		     DmOpenRef DB) {  
  ULong numRecs;
  ULong i;
  UInt cardNo;
  LocalID localID;
  ULong ID;  
  VoidHand rec;
  void * recPtr;
  UInt l;

  DmOpenDatabaseInfo(infoDB, &localID, NULL, NULL, &cardNo, NULL);
  DmDatabaseSize(cardNo, localID, &numRecs, NULL, NULL);
  
  if (numRecs) {
    for (i = 0; i < numRecs; i++) {
      Boolean match = false;
      rec = DmGetRecord(infoDB, i);
      recPtr = MemHandleLock(rec);
      if (*((Word *) (recPtr + sizeof(ULong))) == URLSize) {
	match = !(MemCmp(URL, recPtr+sizeof(ULong)+sizeof(Word)+sizeof(ULong), 
			 (ULong) URLSize));
      } else { 
	match = false;
      }
      ID = *((ULong*) recPtr);
      MemHandleUnlock(rec);
      DmReleaseRecord(infoDB, i, false);      
      if (match) {
	DmFindRecordByID(DB, ID, &l);
	if (DmQueryRecord(DB, l) != hPage) {
	  DmRemoveRecord(infoDB, i);
	  DmRemoveRecord(DB, l);
	}
      break;	  
      }
    }
  }
}

void TrimCache(ULong size) {  
  ULong numRecs;
  ULong l;
  UInt cardNo;
  LocalID localID;
  ULong ID;  
  VoidHand rec;
  void * recPtr;
  ULong curSize;
  UInt i;
  
  DmOpenDatabaseInfo(cacheInfoDB, &localID, NULL, NULL, &cardNo, NULL);
  DmDatabaseSize(cardNo, localID, &numRecs, NULL, NULL);
  
  if (numRecs) {
    for (l = numRecs-1; l >= 0; l--) {
      DmOpenDatabaseInfo(cacheDB, &localID, NULL, NULL, &cardNo, NULL);
      DmDatabaseSize(cardNo, localID, NULL, &curSize, NULL);
      if (curSize <= size) {
	break;
      } 
      rec = DmGetRecord(cacheInfoDB, l);
      recPtr = MemHandleLock(rec);
      ID = *((ULong*) recPtr);
      MemHandleUnlock(rec);
      DmReleaseRecord(cacheInfoDB, l, false);       
      DmFindRecordByID(cacheDB, ID, &i);
      if (DmQueryRecord(cacheDB, i) != hPage) {
	DmRemoveRecord(cacheInfoDB, l);
	DmRemoveRecord(cacheDB, i);
      }
    }
  }
}

void AddToCache(char * URL, Word URLSize, void * data, UInt dataSize, 
		ULong expr, DmOpenRef infoDB, DmOpenRef DB) {
  ULong ID;
  VoidHand rec;
  UInt index;
  void * recPtr;
  Err err;

  RemoveFromCache(URL, URLSize, infoDB, DB);
  if (infoDB == cacheInfoDB) {
    if (dataSize < cacheSize) {
      TrimCache(cacheSize-dataSize);
    } else {
      TrimCache(0);
    }
  }
  index = 0;  
  rec = DmNewRecord(DB, &index, dataSize);
  recPtr = MemHandleLock(rec);
  DmWrite(recPtr, 0, data, dataSize);
  MemHandleUnlock(rec);
  DmReleaseRecord(DB, index, true);
  err = DmRecordInfo(DB, index, NULL, &ID, NULL);
  ErrFatalDisplayIf(err, "Error geting record info");
  index = 0;
  rec = DmNewRecord(infoDB, &index, URLSize + 
		     sizeof(ULong) + sizeof(Word) + sizeof(ULong));
  recPtr = MemHandleLock(rec);
  DmWrite(recPtr, 0, &ID, sizeof(ULong));
  DmWrite(recPtr, sizeof(ULong), &URLSize, sizeof(Word));
  DmWrite(recPtr, sizeof(ULong) + sizeof(Word), &expr, sizeof(ULong));
  DmWrite(recPtr, sizeof(ULong) + sizeof(Word) + sizeof(ULong), URL, URLSize);
  MemHandleUnlock(rec);
  DmReleaseRecord(infoDB, index, true);  
}

void FlushCache(DmOpenRef infoDB, DmOpenRef DB) {
  ULong i;
  UInt cardNo;
  LocalID localID;
  ULong ID;  
  VoidHand rec;
  void * recPtr;
  UInt l;

  DmOpenDatabaseInfo(infoDB, &localID, NULL, NULL, &cardNo, NULL);
  DmDatabaseSize(cardNo, localID, &i, NULL, NULL);
  
  while(i) {
    --i;
    rec = DmGetRecord(infoDB, i);
    recPtr = MemHandleLock(rec);
    ID = *((ULong*) recPtr);
    MemHandleUnlock(rec);
    DmReleaseRecord(infoDB, i, false);      
    if (DmFindRecordByID(DB, ID, &l) != dmErrUniqueIDNotFound) { 
      if (DmQueryRecord(DB, l) != hPage) {
	DmRemoveRecord(infoDB, i);
	DmRemoveRecord(DB, l);
      }
    }
  }
}
