#include <Pilot.h>
#include "globals.h"
#include "callback.h"
#include "hotlist.h"
#include "history.h"
#include "send.h"
#include "utils.h"
#include "editmenu.h"

static Int editItem;

static UInt addtohotlist(CharPtr url, CharPtr name)
{
    VoidHand dataHand;
    VoidPtr dataPtr;
    UInt index = 65535;
    UInt urllen, namelen;

    if (!url) url = "";

    urllen = StrLen(url) + 1;
    namelen = name ? StrLen(name) + 1 : 0;

    dataHand = DmNewRecord(hotlistDB, &index, (ULong) urllen+namelen);
    ErrFatalDisplayIf(!dataHand, "Out of memory!");
    dataPtr = MemHandleLock(dataHand);
    DmWrite(dataPtr, 0, url, urllen);
    if (name) {
	DmWrite(dataPtr, urllen, name, namelen);
    }
    MemHandleUnlock(dataHand);
    DmReleaseRecord(hotlistDB, index, true);
    return index;
}

void AddUrlToHotList(void) {
  VoidPtr urlPtr;

  if (curURL && curURLLength) {
    urlPtr =  MemHandleLock(curURL);
    addtohotlist(urlPtr, NULL);
    MemHandleUnlock(curURL);
  }
}

void HotListDrawRecord(VoidPtr tableP, Word row, Word column, 
			     RectanglePtr bounds) {
  VoidHand recHandle;
  CharPtr recText;
  Int textLen, width;
  Boolean fits;
  UInt recordNum = 0;
  FontID currFont;

  CALLBACK_PROLOGUE

  WinEraseRectangle(bounds, 0); 
  recordNum = (UInt) TblGetItemInt (tableP, row, column);

  currFont = FntSetFont (stdFont);
  
  recHandle = DmGetRecord(hotlistDB, recordNum);
  recText = MemHandleLock(recHandle);
  if (StrLen(recText)+1 < MemHandleSize(recHandle)) {
    recText += StrLen(recText)+1;
  }
  width = bounds->extent.x - 2;
  textLen = StrLen(recText);
  FntCharsInWidth(recText, &width, &textLen, &fits);
  
  WinDrawChars(recText, textLen, bounds->topLeft.x, bounds->topLeft.y);
  
  MemHandleUnlock(recHandle);    
  
  DmReleaseRecord(hotlistDB, recordNum, false);
  
  FntSetFont (currFont);

  CALLBACK_EPILOGUE
}

Int HotListLoadTable(FormPtr frm, Int top)  {
  Word row;
  Word lastRecordNum;
  Word numRows;
  VoidHand recordH;
  TablePtr tableP;
  ULong numRecs;
  LocalID lid;
  UInt cardno;
  Int recordNum;
  

  tableP = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, HotListTable));

  numRows = TblGetNumberOfRows (tableP);

  TblUnhighlightSelection(tableP);

  DmOpenDatabaseInfo(hotlistDB, &lid, NULL, NULL, &cardno, NULL);
  DmDatabaseSize(cardno, lid, &numRecs, NULL, NULL);

  if ((top+numRows) > numRecs) {
    top = numRecs - numRows;
  }
  if (top < 0) {
    top = 0;
  }
  /*
  FrmUpdateScrollers(frm, HotListUp, HotListDown, top, top+numRows 
                     < numRecs);
  */

  recordNum = top;

  for (row = 0; row < numRows; row++, recordNum++) {
    recordH = DmQueryRecord (hotlistDB, recordNum);
    
    if (recordH) {
      TblSetItemStyle (tableP, row, 0, customTableItem);      
      TblSetItemInt (tableP, row, 0, (Word) recordNum);
      TblSetRowUsable (tableP, row, true);
      lastRecordNum = recordNum;
    } else {
      TblSetRowUsable (tableP, row, false);
    }
    TblMarkRowInvalid (tableP, row); 
  }
  return top;
}

void HotListTableLoad(FormPtr frm) {
  TablePtr tableP;

  tableP = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, HotListTable));
  curTop = HotListLoadTable (frm, 0);
  TblSetCustomDrawProcedure (tableP, 0, HotListDrawRecord);
  TblSetColumnUsable (tableP, 0, true);	
}

Boolean EditHotlistURLHandleEvent(EventPtr e) {
  FormPtr frm;
  Boolean handled=false;
  char * url;
  char * name; 
  char * buf;
  VoidHand rec;

  CALLBACK_PROLOGUE

  switch (e->eType) {
  case frmOpenEvent:
    rec = DmQueryRecord(hotlistDB, editItem);
    url = MemHandleLock(rec);
    if (StrLen(url)+1 < MemHandleSize(rec)) {
      name = url + StrLen(url)+1;
    } else {
      name = 0;
    }
    frm = FrmGetActiveForm();
    if (name) {
      SetFld(FrmGetObjectPtr(frm, FrmGetObjectIndex (frm, EditHLURLNameF)), 
	     name, StrLen(name));
    }
    SetFld(FrmGetObjectPtr(frm, FrmGetObjectIndex (frm, EditHLURLURLF)), 
	   url, StrLen(url));
    MemHandleUnlock(rec);
    FrmDrawForm(frm);
    handled = true;
    break;
  case ctlSelectEvent:
    frm = FrmGetActiveForm();  
    switch(e->data.ctlEnter.controlID) {
    case EditHLURLSave:
      name = FldGetTextPtr(FrmGetObjectPtr(frm, 
					FrmGetObjectIndex (frm, 
							   EditHLURLNameF)));
      url = FldGetTextPtr(FrmGetObjectPtr(frm, 
					FrmGetObjectIndex (frm, 
							   EditHLURLURLF)));
      if (name == NULL || StrLen(name) == 0) {
	DmResizeRecord(hotlistDB, editItem, StrLen(url)+1);
	rec = DmQueryRecord(hotlistDB, editItem);
	buf = MemHandleLock(rec);
	DmWrite(buf, 0, url, StrLen(url)+1);
	MemHandleUnlock(rec);
      } else {
	DmResizeRecord(hotlistDB, editItem, StrLen(url)+StrLen(name)+2);
	rec = DmQueryRecord(hotlistDB, editItem);
	buf = MemHandleLock(rec);
	DmWrite(buf, 0, url, StrLen(url)+1);
	DmWrite(buf, StrLen(url)+1, name, StrLen(name)+1);	
	MemHandleUnlock(rec);
      }
    case EditHLURLCancel:
      FrmReturnToForm(0);
    }
    handled = true;
    break;
  case menuEvent:
    MenuEraseStatus(NULL);
    if (EditMenuHandleEvent(e->data.menu.itemID, FrmGetActiveForm())) {
	handled = true;
    } else {
	switch(e->data.menu.itemID) {
	}
    }
    handled = true;
    break;

  default:
    break;
  }    

  CALLBACK_EPILOGUE
    
  return handled;
}

Boolean HotListHandleEvent (EventPtr e) {

  FormPtr frm;
  Boolean handled=false;

  CALLBACK_PROLOGUE

  switch (e->eType) {
  case winEnterEvent:
    frm = FrmGetActiveForm();  
    FrmDrawForm(frm);
    handled = true;
    break;    
  case frmOpenEvent:    
    frm = FrmGetActiveForm();  
    HotListTableLoad(frm);    
    FrmDrawForm(frm);
    handled = true;
    break;
  case keyDownEvent:
    frm = FrmGetActiveForm();  
    if (e->data.keyDown.chr == pageUpChr) {	
      if (curTop) {
	curTop = HotListLoadTable (frm, curTop-2);      
	TblDrawTable( FrmGetObjectPtr 
		      (frm, FrmGetObjectIndex (frm, HotListTable)));
      }
      handled = true;
    }
    else if (e->data.keyDown.chr == pageDownChr) {	
      curTop = HotListLoadTable (frm, curTop+2);      
      TblDrawTable( FrmGetObjectPtr 
		    (frm, FrmGetObjectIndex (frm, HotListTable)));
      handled = true;
    }
    break;
  case ctlSelectEvent:
    frm = FrmGetActiveForm();  
    switch(e->data.ctlEnter.controlID) {
      Word row, col;
      TablePtr tblP; 
      VoidHand loc;
      void * src,* dst;
      VoidHand tcurURL;
      ULong tcurURLLength;
      UInt newpos, newtop;
      Word numRows;

    case HotListGoto:  
      tblP =  FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, HotListTable));
      if (TblGetSelection(tblP, &row, &col)) {

	loc = DmQueryRecord(hotlistDB, TblGetItemInt(tblP, row, col));
	src = MemHandleLock(loc);
	tcurURLLength = StrLen(src)+1;
	tcurURL = MemHandleNew((ULong) tcurURLLength);
	dst = MemHandleLock(tcurURL);
	MemMove(dst, src, tcurURLLength);
	MemHandleUnlock(loc);
	MemHandleUnlock(tcurURL);
	
	if (!SendURL(&conID, tcurURL, true, true, false, true)) {
	  if (tcurURL) {
	    MemHandleFree(tcurURL);
	  }
	}
	DmReleaseRecord(hotlistDB, TblGetItemInt(tblP, row, col), false);
      }
    case HotListCancel:
      FrmGotoForm(MainForm);
      handled = true;
      break;
    case HotListNew:
      tblP =  FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, HotListTable));
      numRows = TblGetNumberOfRows (tblP);
      newpos = addtohotlist("http://", "New entry");
      newtop = curTop;
      if (newpos >= curTop + numRows) {
	newtop = newpos - numRows + 1;
      }
      if (newpos < curTop) {
	newtop = newpos;
      }
      curTop = HotListLoadTable (frm, newtop);      
      TblSelectItem(tblP, newpos - curTop, 0);
      TblDrawTable(tblP);
      /* FALLTHROUGH */
    case HotListEdit:
      tblP =  FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, HotListTable));
      if (TblGetSelection(tblP, &row, &col)) {	
	editItem = TblGetItemInt(tblP, row, col);
	FrmPopupForm(EditHotlistURLForm);
      }
      handled = true;
      break;
    case HotListDelete:
      tblP =  FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, HotListTable));
      if (TblGetSelection(tblP, &row, &col)) {	
	DmRemoveRecord(hotlistDB, TblGetItemInt(tblP, row, col));
	HotListLoadTable(frm, curTop);
	TblUnhighlightSelection(tblP);
	TblDrawTable(tblP);
      }
      handled = true;
      break;
    }
  case ctlEnterEvent:
  case ctlRepeatEvent:
    frm = FrmGetActiveForm();  
    switch(e->data.ctlEnter.controlID) {
    case HotListUp:
      if (curTop) {
	curTop = HotListLoadTable (frm, curTop-2);      
	TblDrawTable( FrmGetObjectPtr 
		      (frm, FrmGetObjectIndex (frm, HotListTable)));
      }
      handled = true;
      break;
    case HotListDown:     
      curTop = HotListLoadTable (frm, curTop+2);      
      TblDrawTable( FrmGetObjectPtr 
		    (frm, FrmGetObjectIndex (frm, HotListTable)));
      handled = true;
      break;
    }
    break;

  default:
    break;
  }

  CALLBACK_EPILOGUE

  return handled;
} 

