#include <Pilot.h>
#include "globals.h"
#include "loaddb.h"
#include "utils.h"
#include "callback.h"
#include "main.h"


#define dbName(p) ((char *) p)
#define dbFlags(p) (*(Word *)(p+0x20))
#define dbVer(p) (*(Word *)(p+0x22))
#define dbType(p) (*(DWord *)(p+0x3c))
#define dbID(p) (*(DWord *)(p+0x40))

static unsigned char *dbp = NULL; /* DataBasePointer */
static Boolean done;

void DatabaseLoad(VoidHand hDBtoLoad) {
  EventType e;
  Err err;

  dbp = MemHandleLock(hDBtoLoad);
  done = false;
  
  if (dbp == NULL) return;
  
  FrmPopupForm(LoadDatabaseForm);  
  while (!done) {
    EvtGetEvent(&e, -1);
    if (! SysHandleEvent (&e))
      if (! MenuHandleEvent (NULL, &e, &err))
	if (! ApplicationHandleEvent (&e))
	  FrmDispatchEvent (&e);
    if (e.eType == appStopEvent) {
      EvtAddEventToQueue(&e);
      done = true;
    }
  }
  FrmReturnToForm(0);
  MemHandleUnlock(hDBtoLoad);
  dbp = NULL;
}


Boolean LoadDatabaseHandleEvent(EventPtr e) {
  Boolean handled = false;
  FormPtr frm;
  LocalID LID = 0;
  Err err;
#if 0
  UInt cardNo = 0;
  DmSearchStateType DBState;
#endif
  char name[33];

  CALLBACK_PROLOGUE

  switch (e->eType) {
  DWord dbtype;
  int resP;
  FieldPtr kindfld;
  int fldset;

  case frmOpenEvent:
    frm = FrmGetActiveForm(); 
   
    name[33] = 0;
    MemMove(name, dbp, 32);    

    LID = DmFindDatabase(0, name);

    dbtype = dbType(dbp);
    resP = dbFlags(dbp) & ~0x08; /* Clear the BACKUP flag */
    if ((resP != 0 && resP != 1) || resP != dbVer(dbp)) {
	/* This is suspicious... */
	resP = 0;
    }
    
    /* SETUP FIELDS */

    SetFld(FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, LoadDBNameF)),	   
	   name, 33);

    fldset = 0;
    kindfld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, LoadDBKindF));
    if (resP) {
	switch(dbtype) {
	case 'appl':
	    SetFld(kindfld, "This seems to be an application.", 32);
	    fldset = 1;
	    break;
	case 'HACK':
	    SetFld(kindfld, "This seems to be a Hackmaster hack.", 35);
	    fldset = 1;
	    break;
	case 'GLib':
	    SetFld(kindfld, "This seems to be a GLib shared library.", 39);
	    fldset = 1;
	    break;
	case 'libr':
	    SetFld(kindfld, "This seems to be a SysLib shared library.", 41);
	    fldset = 1;
	    break;
	}
    } else {
	if (dbtype == 'TEXt' && dbID(dbp) == 'REAd') {
	    SetFld(kindfld, "This seems to be an AportisDoc file.", 36);
	    fldset = 1;
	}
	if (dbtype == 'vIMG' && dbID(dbp) == 'View') {
	    SetFld(kindfld, "This seems to be an Image Viewer file.", 38);
	    fldset = 1;
	}
    }
    if (!fldset) {
	SetFld(kindfld, "This is an unknown database type.", 33);
    }

    SetFld(FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, LoadDBTypeF)),
	   (char *) &dbType(dbp), 4);
    SetFld(FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, LoadDBIDF)),
	   (char *) &dbID(dbp), 4);
#if 0
    if (!LID) {
      DmGetNextDatabaseByTypeCreator(true, &DBState, dbType(dbp),
				     dbID(dbp), false, &cardNo, &LID);
    }
#endif
    if (LID) {
      SetFld(FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, LoadDBAlready)),
	     "THIS WILL REPLACE AN EXISTING DATABASE.", 39);
    } else {
      SetFld(FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, LoadDBAlready)),
	     "This will not replace an existing database.", 43);
    }

    FrmDrawForm(frm);
    
    handled = true;
    break;

  case ctlSelectEvent:    
    switch(e->data.ctlEnter.controlID) {
    case LoadDBYes:
      /* SAVE DATABASE */
      name[33] = 0;
      MemMove(name, dbp, 32);
      do {
	err = 0;
	LID = DmFindDatabase(0, name);      
	if (LID) {
	  err = DmDeleteDatabase(0, LID);
	}
#if 0
	if (!err && !DmGetNextDatabaseByTypeCreator(true, &DBState,
	    dbType(dbp), dbID(dbp), false, &cardNo, &LID)) {
	  err = DmDeleteDatabase(cardNo, LID);
	}
#endif
	if (LID && err) {
	    /* Badness.  We couldn't delete the database */
	    FrmAlert(LoadDBBadnessAlert);
	    goto badness_label;
	} else {
	    LID = 0;
	}
      } while (LID != 0);
      DmCreateDatabaseFromImage(dbp);
badness_label: ;
      /* FALLTHROUGH */

    case LoadDBNo:
      handled = true;
      done = true;
      break;
    }
  default:
  }
    
  CALLBACK_EPILOGUE

  return handled;

}
