/* UI and Data Processing Code for Top Gun Wing Man
   (a proxy based WWW browser for the USRobotics Pilots)
   See the README file for copyright/licensing terms */

#include <Pilot.h>
#include "globals.h"
#include "start.h"
#include "stop.h"
#include "prefs.h"
#include "cache.h"
#include "history.h"
#include "mainform.h"
#include "urlform.h"
#include "aggform.h"
#include "hotlist.h"
#include "pagedisplay.h"
#include "offlineform.h"
#include "splashform.h"
#include "metadata.h"
#include "notify.h"
#include "loaddb.h"
#include "main.h"

Boolean ApplicationHandleEvent(EventPtr e)
{
  FormPtr frm;
  Word    formId;
  Boolean handled = false;
  
  switch (e->eType) {

  case frmLoadEvent:
    formId = e->data.frmLoad.formID;
    frm = FrmInitForm(formId);
    FrmSetActiveForm(frm);    

    switch(formId) {
    case MainForm:
      FrmSetEventHandler(frm, MainFormHandleEvent);
      break;
    case PrefsForm:
      FrmSetEventHandler(frm, PrefsFormHandleEvent);
      break;
    case SplashForm:
      FrmSetEventHandler(frm, SplashFormHandleEvent);
      break;
    case URLForm:
      FrmSetEventHandler(frm, URLFormHandleEvent);
      break;
    case AggForm:
      FrmSetEventHandler(frm, AggFormHandleEvent);
      break;
    case HotListForm:
      FrmSetEventHandler(frm, HotListHandleEvent);
      break;
    case OfflineCacheForm:
      FrmSetEventHandler(frm, OfflineCacheHandleEvent);
      break;
    case LoadDatabaseForm:
      FrmSetEventHandler(frm, LoadDatabaseHandleEvent);
      break;      
    case EditHotlistURLForm:
      FrmSetEventHandler(frm, EditHotlistURLHandleEvent);
      break;      
    case EditOLCacheForm:
      FrmSetEventHandler(frm, EditOLCacheHandleEvent);
      break;      
    }
    handled = true;
    break;
  
  case transNewChunk: 
    {      
      ChunkPtr c;      
      transEventPtr te = (transEventPtr) e;
      
      frm = FrmGetActiveForm();
      if (curChunk) {
	Err err;
	err = DestroyChunk(curChunk);
      }
      
      if (hPage) {
	  hPage = NULL;
	  UpdateHiLiPiece(-2, NULL, 0, 0);
	  ScrollReset();
      }
      curChunk = te->data.newChunk.chunkHand;
      
      c = (ChunkPtr) MemHandleLock(curChunk);
      if (c->requestID == currRequestID || c->requestID & 0x8000000) {
        CheckAggList(c->metaHand);
        switch(GetPageType(c->metaHand)) {
	VoidHand hDBtoLoad;

        case PageTypeRegular:
	    hPage = c->dataHand;
	    UpdateHiLiPiece(-2, NULL, 0, 0);
	    ScrollReset();
	    MemHandleUnlock(curChunk);
	    if (hPage && transURL) {
	      char * url;
	      VoidHand data;
		    
	      if (curURL) {
		MemHandleFree(curURL);
	      }
	      
	      curURL = transURL;
	      curURLLength = transURLLength;
	      
	      if (tcurHistory != -1) {
		curHistory = tcurHistory;			
		tcurHistory = -1;	
	      }

	      PutInHistory(curURL, curURLLength);
	      
	      transURL = NULL;
	      transURLLength = 0;

	      url = MemHandleLock(curURL);
	      data = InCache(url, curURLLength, offlineInfoDB, offlineDB);
	      if (data) {
		AddToCache(url, curURLLength, 
			   (char *) MemHandleLock(hPage), c->dataSize, 0,
			   offlineInfoDB, offlineDB);
	      } else {
		AddToCache(url, curURLLength, 
			   (char *) MemHandleLock(hPage), c->dataSize, 
			   ExprTime(c->metaHand), cacheInfoDB, cacheDB);
	      }
	      MemHandleUnlock(hPage);
	      MemHandleUnlock(curURL);
	    } 
	    break;

	case PageTypeDatabase:
	    /* Save the given database */
	    hDBtoLoad = c->dataHand;
	    hPage = NULL;

	    /* Do something useful with hDBtoLoad */
	    DatabaseLoad(hDBtoLoad);

	    /* Now we probably want to reshow the current URL (though we'll
	       jump to the top...) */
	    if ((curHistory == -1) || MoveInHistory(0)) {
		EraseMainWindow();
		ScrollReset();
		if (hPage) {
		    WinDrawProxyString(hPage, currXPosOffset, currYPosOffset);
		}
	    }
	    break;

	default:
	    break;
	}
	handled = true;
      }
      if (hPage && (frm->formId == MainForm)) {
        EraseMainWindow();
	ScrollReset();
        WinDrawProxyString(hPage, currXPosOffset, currYPosOffset); 
      }
      break;
    }
  default:
    break;
  }
  return handled;
}


static Boolean TopLevelHandleEvent(EventPtr e) {
  FormPtr frm;
  Boolean handled = false;

  frm = FrmGetActiveForm();
  if (e->eType == keyDownEvent) {
    switch (e->data.keyDown.chr) {
    case hard2Chr: 
    case hard3Chr: 
      if (remap) {
	handled = true;
	if (frm->formId == MainForm) {
	  if (MoveInHistory(e->data.keyDown.chr == hard2Chr ? 1 : -1)) { 
	    EraseMainWindow();
	    ScrollReset();
	    if (hPage) {
	      WinDrawProxyString(hPage, currXPosOffset, currYPosOffset);
	    }
	  }
	}
      }
      break;
    default:
      break;
    }      
  }
  return handled;
}

/* The main event loop */
static void EventLoop(void)
{
  Word err;
  EventType e;
  do {
    if (conID != -1) {
      CheckNetwork();
    }
    EvtGetEvent(&e, 200); /* why does -1 work anyway? */
    if (!TopLevelHandleEvent(&e))
      if (! SysHandleEvent (&e))
	if (! MenuHandleEvent (NULL, &e, &err))
	  if (! ApplicationHandleEvent (&e))
	    FrmDispatchEvent (&e);
  } while (e.eType != appStopEvent);
}

/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
  Word err;
  if (cmd == sysAppLaunchCmdNormalLaunch) {
    err = StartApplication();    
    if (err) return err;
    EventLoop();
    StopApplication();

  } else {
    return sysErrParamErr;
  }

  return 0;
}

