#include <Pilot.h>
#include "globals.h"
#include "callback.h"
#include "mainform.h"
#include "pagedisplay.h"
#include "history.h"
#include "hotlist.h"
#include "cache.h"
#include "send.h"
#include "notify.h"

Boolean MainFormHandleEvent (EventPtr e) {
  Boolean handled = false;
  FormPtr frm;
  int sendRes;
  static int dragX = -1;
  static int dragY = -1;

  CALLBACK_PROLOGUE
  
  switch (e->eType) {
  case keyDownEvent:
    if (e->data.keyDown.chr == pageUpChr) {	
      ScrollPage(0, -scrollSize);
      handled = true;
    } else if (e->data.keyDown.chr == pageDownChr) {
      ScrollPage(0, scrollSize);
      handled = true;
    }
    break;

  case frmCloseEvent:
    EraseMainWindow();
    handled = false;
    break;

  case frmOpenEvent:
    frm = FrmGetActiveForm();
    /* I'm going to try getting rid of this; let me know if there
       are droppings left over from something... - Ian */
    /* WinEraseWindow(); */
    FrmDrawForm(frm);
    StatusDraw(0);
    ScrollRefresh();
    if (hPage) {
      WinDrawProxyString(hPage, currXPosOffset, currYPosOffset);
    }
    handled = true;
    break;

  case menuEvent:
    MenuEraseStatus(NULL);

    switch(e->data.menu.itemID) {
      char * url;
      void * data;
      VoidHand tcurURL;
      ULong tcurURLLength;

    case GetMenu: 
      FrmPopupForm(URLForm);
      break;

    case AggregatorMenu:
      FrmPopupForm(AggForm);
      break;

    case StopMenu:
      cancelSend(conID);
      break;      

    case BackMenu:
      if (MoveInHistory(1)) { 
	EraseMainWindow();
	ScrollReset();
	if (hPage) {
	  WinDrawProxyString(hPage, currXPosOffset, currYPosOffset);
	}
      }
      break;
    case ForwardMenu:
      if (MoveInHistory(-1)) {
	EraseMainWindow();		  
	ScrollReset();
	if (hPage) {
	  WinDrawProxyString(hPage, currXPosOffset, currYPosOffset);
	}
      }
      break;
    case SourceMenu:
    case DocMenu:
      if (curURL) {
	char *t, *c;
	int isdoc = (e->data.menu.itemID == DocMenu);

	c = MemHandleLock(curURL);
	/* View source:
	     o If curURL starts with src: or doc:, remove it
	     o Otherwise, prepend src:
	   View source as Doc:
	     o If curURL starts with src: or doc:, remove it
	     o _As well_, prepend doc:
	*/
	if (curURLLength >= 4 && ((c[0] == 's' && c[1] == 'r') ||
	    (c[0] == 'd' && c[1] == 'o')) && c[2] == 'c' && c[3] == ':') {
	    tcurURLLength = curURLLength - (isdoc ? 0 : 4);
	    tcurURL = MemHandleNew(tcurURLLength);
	    t = MemHandleLock(tcurURL);
	    if (isdoc) {
		MemMove(t, "doc:", 4);
	    }
	    MemMove(t + (isdoc ? 4 : 0), c+4, tcurURLLength);
	} else {
	    tcurURLLength = curURLLength + 4;
	    tcurURL = MemHandleNew(tcurURLLength);
	    t = MemHandleLock(tcurURL);
	    MemMove(t, isdoc ? "doc:" : "src:", 4);
	    MemMove(t+4, c, curURLLength);
	}
	
	MemHandleUnlock(curURL);
	MemHandleUnlock(tcurURL);

	if (!SendURL(&conID, tcurURL, true, true, false, true)) {
	  if (tcurURL) {
	    MemHandleFree(tcurURL);
	  }
	}
      }
      break;
    case PrefsMenu:
      FrmGotoForm(PrefsForm);
      break;
    case HotListMenu:
      FrmGotoForm(HotListForm);
      break;    
    case AddHotListMenu:
      AddUrlToHotList();
      break;
    case ReloadMenu: 
      if (curURL) {
	void *t, *c;
	
	if (transURL) {
	  MemHandleFree(transURL);
	}
	transURLLength = curURLLength;
	transURL = MemHandleNew(transURLLength);
	
	t = MemHandleLock(transURL);
	c = MemHandleLock(curURL);

	MemMove(t, c, transURLLength);
	
	MemHandleUnlock(curURL);
	MemHandleUnlock(transURL);
	
	SendURL(&conID, transURL, false, false, true, false);

      }
      break;
    case AddCacheMenu:
      if (hPage && curURL) {
	url = (char *) MemHandleLock(curURL); 
	data = MemHandleLock(hPage); 
	AddToCache(url, MemHandleSize(curURL), data, MemHandleSize(hPage),
		   0, offlineInfoDB, offlineDB);
	MemHandleUnlock(hPage);
	MemHandleUnlock(curURL);
      }
      break;
    case EditCacheMenu:
      FrmGotoForm(OfflineCacheForm);
      break;
    case FlushCacheMenu:
      FlushCache(cacheInfoDB, cacheDB);
      break;
    case InfoMenu:
      tcurURLLength = StrLen(INFO_URL) + 1;
      tcurURL = MemHandleNew((ULong)tcurURLLength);
      if (tcurURL) {
	char *urltxt = MemHandleLock(tcurURL);
        StrCopy(urltxt, INFO_URL);
        MemHandleUnlock(tcurURL);
      } else {
	ErrFatalDisplayIf(tcurURLLength, "Out of memory!");
      }
      sendRes = SendURL(&conID, tcurURL, true, true, false, true);
      if (!sendRes) {
	if (tcurURL) {
	  MemHandleFree(tcurURL);
	}
      }
      break;
    }

    handled = true;
    break;    
  case penDownEvent:
    if (hPage) {
      char *outURL;
      int eventPieceOff;
      SWord xcoord, ycoord;
      dragX = -1;
      dragY = -1;
      if (ScrollHandleEvent(e)) {
        handled = true;
      } else {
	eventPieceOff = PieceClicked(hPage, e->screenX, e->screenY, 
				     currXPosOffset, currYPosOffset,
				     &xcoord, &ycoord);
	if (eventPieceOff >= 0) {
	  int imgClkResult;
	  SWord *hPtr, *eventPiece;

	  hPtr = MemHandleLock(hPage);
	  ErrFatalDisplayIf(!hPtr, "Cannot lock hPage");
	  eventPiece = hPtr + eventPieceOff;

	  imgClkResult = ClickImage(e, eventPiece, currXPosOffset,
					currYPosOffset);

	  MemHandleUnlock(hPage);

	  if (imgClkResult == -1) {
	    if (UpdateHiLiPiece(-1, eventPieceOff, currXPosOffset,
		currYPosOffset)) {
		handled = true;
	    }
	  } else if (imgClkResult == 0) {
	    handled = true;
	  } else {
	    /* This is an IMAGE */
	    VoidHand tcurURL;
	    ULong tcurURLLength;

	    hPtr = MemHandleLock(hPage);
	    ErrFatalDisplayIf(!hPtr, "Cannot lock hPage");
	    eventPiece = hPtr + eventPieceOff;

	    if (imgClkResult == 1) { /* if sending the first URL */
	      int ismap = eventPiece[6] & IMAGE_ISMAP;
	      int hasserv = eventPiece[6] & IMAGE_HAS_SERV_SIZE;
	      char coordbuf[15];

	      /* create buffer for sending info */
	      tcurURLLength = eventPiece[hasserv ? 9 : 7];
	      if (ismap) {
		  SWord cli_x = eventPiece[4];
		  SWord cli_y = eventPiece[5];
		  SWord serv_x, serv_y;
		  if (hasserv) {
		    serv_x = eventPiece[7]; serv_y = eventPiece[8];
		  } else {
		    serv_x = cli_x; serv_y = cli_y;
		  }
		  if (!cli_x) cli_x = 1;
		  if (!cli_y) cli_y = 1;
		  /* %d can be at most 6 chars ("-32768") */
		  StrPrintF(coordbuf, "?%d,%d",
			     (int)((Long)xcoord * (Long)serv_x / (Long)cli_x),
			     (int)((Long)ycoord * (Long)serv_y / (Long)cli_y));
		  tcurURLLength += StrLen(coordbuf);
	      }
	      tcurURL = MemHandleNew((ULong) tcurURLLength);
	      if (tcurURL) {
		outURL = MemHandleLock(tcurURL);
		/* data out setup */
		StrCopy (outURL, (char*) &eventPiece[hasserv ? 10 : 8]);
		if (ismap) {
		    StrCat(outURL, coordbuf);
		}
		MemHandleUnlock(tcurURL);
	      } else {
		ErrFatalDisplayIf (tcurURLLength, "Not enough memory!");
	      }
	    } else if (imgClkResult == 2) { /* if sending second URL */
	      int hasserv = eventPiece[6] & IMAGE_HAS_SERV_SIZE;

	      /* skip past first URL */
	      int index = hasserv ? 9 : 7;
	      tcurURLLength = eventPiece[index++];	    
	      if ((tcurURLLength % 2) == 0) {
		index += (tcurURLLength >> 1);
	      } else {
		index += ((tcurURLLength + 1) >> 1);
	      }
	      /* create buffer for sending info */
	      tcurURLLength = eventPiece[index++];
	      tcurURL = MemHandleNew((ULong) tcurURLLength);
	      if (tcurURL) {
		outURL = MemHandleLock(tcurURL);
		/* data out setup */
		StrCopy (outURL, (char *) &eventPiece[index]);
		MemHandleUnlock(tcurURL);
	      } else {
		ErrFatalDisplayIf (tcurURLLength, "Not enough memory!");
	      }
	    } else {
	      tcurURLLength = 0;
	      tcurURL = 0;
	      ErrFatalDisplayIf (tcurURLLength, "Internal Error");
	    }

	    MemHandleUnlock(hPage);

	    /* send out request */	    
	    sendRes = SendURL(&conID, tcurURL, true, true, false, true);
	    if (!sendRes) {
	      MemHandleFree (tcurURL);
	    }

	    handled = true;
	  }
	}
	if (!handled && e->screenX < MAIN_DISPLAY_WIDTH &&
		    e->screenY < MAIN_DISPLAY_HEIGHT) {
	    /* A penDown in the main screen that didn't do anything else */
	    dragX = e->screenX;
	    dragY = e->screenY;
	    handled = true;
	}
      } /* end drag test else */
    }
  break;
  
  case penMoveEvent:
    {
      if (hPage) {
	if (ScrollHandleEvent(e)) {
	  handled = true;
	} else if (dragX >= 0) {
	    int diffx, diffy, sx, sy;
	    sx = e->screenX;
	    sy = e->screenY;
	    if (sx >= MAIN_DISPLAY_WIDTH) sx = MAIN_DISPLAY_WIDTH - 1;
	    if (sy >= MAIN_DISPLAY_HEIGHT) sy = MAIN_DISPLAY_HEIGHT - 1;
	    diffx = sx - dragX;
	    diffy = sy - dragY;
	    if (((diffx > 0) ? diffx : -diffx) > SCROLL_THRESH ||
	        ((diffy > 0) ? diffy : -diffy) > SCROLL_THRESH) {
		ScrollPage(-diffx, -diffy);
		dragX = sx;
		dragY = sy;
	    }
	} else {
	  /* standard link & image click handler */
	  int eventPieceOff = PieceClicked(hPage, e->screenX, e->screenY, 
					   currXPosOffset, currYPosOffset,
					   NULL, NULL);
	  /* no event piece chk because it must track movement off piece */
	  UpdateHiLiPiece(0, eventPieceOff, currXPosOffset, currYPosOffset);
	  handled = true;
	}
      }
    }
  break;

  case penUpEvent:
    if (hPage) {
      if (ScrollHandleEvent(e)) {
        handled = true;
      } else if (dragX >= 0) {
	    int diffx, diffy, sx, sy;
	    sx = e->screenX;
	    sy = e->screenY;
	    if (sx >= MAIN_DISPLAY_WIDTH) sx = MAIN_DISPLAY_WIDTH - 1;
	    if (sy >= MAIN_DISPLAY_HEIGHT) sy = MAIN_DISPLAY_HEIGHT - 1;
	    diffx = sx - dragX;
	    diffy = sy - dragY;
	    ScrollPage(-diffx, -diffy);
	    dragX = -1;
	    dragY = -1;
      } else {
	char *outURL;
	/* standard link & image click handler */
	int eventPieceOff = PieceClicked(hPage, e->screenX, e->screenY, 
					 currXPosOffset, currYPosOffset,
					 NULL, NULL);
	if (eventPieceOff >= 0) {
	  SWord *hPtr, *eventPiece;
	  Word attributes;
	  
	  hPtr = MemHandleLock(hPage);
	  ErrFatalDisplayIf(!hPtr, "Cannot lock hPage");
	  eventPiece = hPtr + eventPieceOff;
	  
	  attributes = eventPiece[6];
	  
	  MemHandleUnlock(hPage);
	  
	  if (UpdateHiLiPiece(1, eventPieceOff, currXPosOffset,
				currYPosOffset) && (attributes & LINK)
				&& !(attributes & IMAGE)) {
	    /* create buffer for sending info */
	    ULong tcurURLLength;
	    VoidHand tcurURL;
	    
	    hPtr = MemHandleLock(hPage);
	    ErrFatalDisplayIf(!hPtr, "Cannot lock hPage");
	    eventPiece = hPtr + eventPieceOff;
	    
	    tcurURLLength = eventPiece[7];
	    tcurURL = MemHandleNew((ULong) tcurURLLength);
	    if (tcurURL) {
	      outURL = MemHandleLock(tcurURL);
	      /* data out setup */
	      StrCopy (outURL, (char*) &eventPiece[8]);
	      MemHandleUnlock(tcurURL);
	    } else {
	      ErrFatalDisplayIf (tcurURLLength, "Not enough memory!");
	    }
	    
	    MemHandleUnlock(hPage);
	    
	    /* set out link request */
	    sendRes = SendURL(&conID, tcurURL, true, true, false, true);
	    if (!sendRes) {
	      if (tcurURL) {
		MemHandleFree(tcurURL);
	      }		
	    }
	    /* cleanup */	    
	    handled = true;
	  }
	}
      } /* end link drag test */
    }
    break;  

  case winEnterEvent:
    /* Refresh the screen */
    if (e->data.winEnter.enterWindow == FrmGetWindowHandle(FrmGetActiveForm()) &&
        e->data.winEnter.enterWindow !=	e->data.winEnter.exitWindow) {
	if (hPage) {
	  WinEraseWindow();
	  ScrollRefresh();
	  WinDrawProxyString(hPage, currXPosOffset, currYPosOffset);	      
	}
    }
    StatusDraw(0);
    break;

  default:
    break;
  }


  CALLBACK_EPILOGUE

  return handled;
}

