#include <Pilot.h>
#include "globals.h"
#include "metadata.h"
#include "aggform.h"

/* 
 * THE FORMAT OF THE METADATA IS...
 * WORD TYPE, WORD SIZE, SIZE*BYTE DATA...
 * IN NETWORK ORDER OF COURSE.
 */

/* 
 * trys to find the metaData tag with type type in the metadata data.
 * if it put that size in size and return a pointer to the start
 * of the data (if size is non NULL).
 */
static void * FindType(void * data, ULong dataSize, MetaType type, 
		       Word *size) {
  Word t;
  Word s;
  Word seen;

  seen = 0;  
  while (seen+(2*sizeof(Word)) <= dataSize) {    
    t =  *((Word *) ((void *) data+seen));
    s = *((Word *) ((void *) data+seen+sizeof(Word)) );
    
    if ((t == type) && (seen+s+2*sizeof(Word) <= dataSize)) {
      if (size) {*size = s;}
      return (data + seen + 2*sizeof(Word));
    }
    seen += s + 2*sizeof(Word);
  }
    
  if (size) {*size = 0;}
  return NULL;
}

/* 
 * returns expr time for the page.  If no expr time returns 0.
 */
ULong ExprTime(VoidHand metaData) {
  ULong dataSize;
  ULong exprtime = 0;
  Word size;
  void * data;
  void * info;
  
  if (metaData) {
    data = MemHandleLock(metaData);
    dataSize = MemHandleSize(metaData);
    
    if ((info = FindType(data, dataSize, MetaCacheExpr, &size)) != NULL) {
	exprtime = *((ULong *)info);
    }

    MemHandleUnlock(metaData);
  }

  return exprtime;
}

/* 
 * returns page type for the page
 */
PageType GetPageType(VoidHand metaData) {
  ULong dataSize;
  PageType type = PageTypeRegular;
  Word size;
  void * data;
  void * info;
  
  if (metaData) {
    data = MemHandleLock(metaData);
    dataSize = MemHandleSize(metaData);
    
    if ((info = FindType(data, dataSize, MetaPageType, &size)) != NULL) {
	type = *((Word *)info);
    }

    MemHandleUnlock(metaData);
  }

  return type;
}

/* 
 * check to see if there's a new agglist in the metadata
 */
void CheckAggList(VoidHand metaData) {
  ULong dataSize;
  Word size;
  void * data;
  void * info;
  
  if (metaData) {
    data = MemHandleLock(metaData);
    dataSize = MemHandleSize(metaData);
    
    if ((info = FindType(data, dataSize, MetaAggList, &size)) != NULL) {
        /* We have a new agg list; install it */
	GotAggList(info, size);
    }

    MemHandleUnlock(metaData);
  }
}

/* Construct a handle to a newly-created metadata block, consisting of
   "num" fields, with values taken from the array "fields".  The length of
   the result is stored in *rlen */
VoidHand MakeMetaData(MetaField *fields, int num, DWord *rlen)
{
    int i;
    DWord totlen = 0;
    VoidHand hnd;
    CharPtr metadata, curptr;

    /* Calculate the total length of the metadata */
    for(i=0;i<num;++i) {
	totlen += fields[i].len + 2*sizeof(Word);
    }
    if (rlen) *rlen = totlen;

    /* Bail early if there's nothing to do */
    if (totlen == 0) {
	return NULL;
    }

    /* Allocate that amount */
    hnd = MemHandleNew(totlen);
    if (!hnd) {
	/* Bail; *rlen != 0 && hnd == NULL implies error */
	return NULL;
    }

    metadata = MemHandleLock(hnd);
    if (!metadata) {
	MemHandleFree(hnd);
	return NULL;
    }

    /* Construct the metadata */
    curptr = metadata;
    for (i=0;i<num;++i) {
	*(Word *)curptr = fields[i].type; curptr += 2;
	*(Word *)curptr = fields[i].len; curptr += 2;
	if (fields[i].len) MemMove(curptr, fields[i].data, fields[i].len);
	curptr += fields[i].len;
    }

    /* That's all */
    MemHandleUnlock(hnd);
    return hnd;
}
