#include <Pilot.h>
#include "trans.h"
#include "globals.h"
#include "notify.h"
#include "pagedisplay.h"
#include "2bit.h"

static DataSize total;
static DataSize seen;
static WinHandle statbarbufWin = NULL;

#define STATBAR_WIDTH 160
#define STATBAR_HEIGHT (157-MAIN_DISPLAY_HEIGHT)
#define STATBAR_X 0
#define STATBAR_Y (MAIN_DISPLAY_HEIGHT+3)

Err StatusInit(void)
{
    Err err;

    statbarbufWin = WinCreateOffscreenWindow(STATBAR_WIDTH, STATBAR_HEIGHT,
	screenFormat, &err);
    return err;
}

void StatusCleanup(void)
{
    if (statbarbufWin) WinDeleteWindow(statbarbufWin, false);
}

void Notify (DataSize *total, DataSize *seen, void *ptr) {
  RectangleType offrect = { { 0, 0 }, { STATBAR_WIDTH, STATBAR_HEIGHT } };
  FormPtr frm = FrmGetActiveForm();
  WinHandle drawwin;
  FontID oldfont;

  if (!frm || frm->formId != MainForm) return;
 
#ifdef OLD_STATUS_BAR
  if ((total && *total && seen && ((*seen) < (*total))) &&
      conID != -1) {
      WinDrawRectangle(&r, 0);
      r.extent.x = ((160 * (*seen)) / (*total)) +1;
      WinFillRectangle(&r, 0);
  } else {
      WinEraseRectangle(&r, 0);
  }
#else
  drawwin = WinGetDrawWindow();
  WinSetDrawWindow(statbarbufWin);
  WinEraseWindow();
  oldfont = FntGetFont();
  FntSetFont(stdFont);
  if ((total && *total && seen && ((*seen) < (*total))) &&
      conID != -1) {
      if ((*seen) > 0) {
        RectangleType r = { { 45, 1 }, { 0, STATBAR_HEIGHT-2 } };
	static char loading[11] = { "Loaded 00%" };
	Word p = ((100 * (DWord)(*seen)) / (DWord)(*total));
	if (p > 99) p = 99;
	loading[7] = (p>=10)?('0'+(p/10)):' ';
	loading[8] = '0'+(p%10);
	WinDrawChars(loading, 10,
	    STATBAR_WIDTH/2-(FntCharsWidth(loading,10)/2), 0);
	r.extent.x = ((70 * (DWord)(*seen)) / (DWord)(*total));
	WinInvertRectangle(&r, 0);
      } else {
        static char *connecting = "Connecting...";
	WinDrawChars(connecting, 13,
	    STATBAR_WIDTH/2-(FntCharsWidth(connecting,13)/2), 0);
      }
  }
  FntSetFont(oldfont);
  WinSetDrawWindow(drawwin);
  WinCopyRectangle(statbarbufWin, drawwin, &offrect,
		    STATBAR_X, STATBAR_Y, scrCopy);
#endif
}

void ReNotify(void)
{
    Notify(&total, &seen, NULL);
}

void RegisterNotify(ConID conID) {
  RegisterNotifyCallback(conID, Notify, &total, &seen, NULL);
}

/* Draw the status bar */
void StatusDraw(int reset)
{
    FormPtr frm = FrmGetActiveForm();

    if (reset) {
	total = 0;
	seen = 0;
    }

    if (!frm || frm->formId != MainForm) return;

    /* Draw any decorations on the page (dividing lines, etc.) */
    WinDrawLine(MAIN_DISPLAY_WIDTH, MAIN_DISPLAY_HEIGHT, MAIN_DISPLAY_WIDTH+2,
		MAIN_DISPLAY_HEIGHT+2);
    WinDrawLine(MAIN_DISPLAY_WIDTH, MAIN_DISPLAY_HEIGHT+2, MAIN_DISPLAY_WIDTH+2,
		MAIN_DISPLAY_HEIGHT);
    ReNotify();
}

/* Draw text in the status bar field; "" means erase field;
   NULL means restore field.
   
   crunchtype specifies how to crunch strings that are too long:
   0: show the center of the string (default)
   1: the string is a URL; do something sane
*/
void StatusBarText(unsigned char *text, int crunchtype)
{
    RectangleType r;
    int l;

    r.topLeft.x = STATBAR_X;
    r.topLeft.y = STATBAR_Y;
    r.extent.x = STATBAR_WIDTH;
    r.extent.y = STATBAR_HEIGHT;

    if (text) {
	l = StrLen(text);
	WinEraseRectangle(&r, 0);
	if (l) {
	    unsigned char *copy = NULL;
	    int width = FntCharsWidth(text,l);
	    int pos = 0;
	    unsigned char *slash;
	    if (width > STATBAR_WIDTH &&
		(copy = MemPtrNew(l+1))) { /* Yes, I mean = */
		switch(crunchtype) {
		case 1:
		    StrCopy(copy, text);
		    text = copy;
		    /* Try to strip http:// */
		    
		    /* changed this to save, oh, umm, umm, 28 bytes of 
		       code space.  also makes it prettier code.
		       If my change doesn't work bash me over the head
		       or something
		    if (text[0] == 'h' && text[1] == 't' && text[2] == 't' &&
			text[3] == 'p' && text[4] == ':' && text[5] == '/' &&
			text[6] == '/') {
		    */
		    if (StrStr(text, "http://") == (CharPtr) text) {
			text += 7;
			l -= 7;
			width = FntCharsWidth(text, l);
			if (width <= STATBAR_WIDTH) break;
		    }

		    /* Draw the hostname/path/to/file; we'll overwrite
		       the end of that later */
		    WinDrawChars(text, l, STATBAR_X, STATBAR_Y);

		    /* Now find the / before the end (but not right at the
		       end), and put '...' before it */
		    for(slash = text+l-2; slash >= text; --slash) {
			if (*slash == '/') break;
		    }
		    if (slash > text) --slash;
		    if (slash < text) ++slash;
		    *slash = '\x85'; /* '\x85' is an ellipsis */
		    l -= (slash - text);
		    text = slash;
		    width = FntCharsWidth(text, l);
		    pos = STATBAR_WIDTH-width;
		    break;

		default:
		    /* Center things */
		    pos = STATBAR_WIDTH/2-width/2;
		    break;
		}
	    }
	    WinDrawChars(text, l, pos, STATBAR_Y);
	    if (copy) MemPtrFree(copy);
	}
    } else {
	ReNotify();
    }
}

typedef enum { SCROLL_ORIENT_VERT, SCROLL_ORIENT_HORIZ } ScrollOrient;

typedef struct {
    int x;
    int y;
    int pagesize;
    int barlength;
    ScrollOrient hv;
    int len;
    int pos;
    int curleft;
    int curright;
} ScrollInfo;

static ScrollInfo hBar = { 0, MAIN_DISPLAY_HEIGHT, MAIN_DISPLAY_WIDTH,
			    MAIN_DISPLAY_WIDTH-1, SCROLL_ORIENT_HORIZ,
			    -1, 0, 0, MAIN_DISPLAY_WIDTH };
static ScrollInfo vBar = { MAIN_DISPLAY_WIDTH, 0, MAIN_DISPLAY_HEIGHT,
                            MAIN_DISPLAY_HEIGHT-1, SCROLL_ORIENT_VERT,
			    -1, 0, 0, MAIN_DISPLAY_HEIGHT };

static void InitInfo(ScrollInfo *b, int len)
{
    b->len = len;
    b->pos = 0;
    b->curleft = 0;
    b->curright = len > b->pagesize ? (ULong)b->pagesize *
			(ULong)b->barlength / (ULong)b->len : b->barlength;
}

static void ScrollUpdate(int newleft, int newright, ScrollInfo *b, int force)
{
    /* Just be silly for now */
    if (force || b->curleft != newleft || b->curright != newright) {
	if (b->hv == SCROLL_ORIENT_VERT) {
	    WinEraseLine(b->x, b->y+b->curleft, b->x, b->y+b->curright);
	    WinEraseLine(b->x+2, b->y+b->curleft, b->x+2, b->y+b->curright);
	    WinDrawLine(b->x, b->y+newleft, b->x, b->y+newright);
	    WinDrawLine(b->x+2, b->y+newleft, b->x+2, b->y+newright);
	} else {
	    WinEraseLine(b->x+b->curleft, b->y, b->x+b->curright, b->y);
	    WinEraseLine(b->x+b->curleft, b->y+2, b->x+b->curright, b->y+2);
	    WinDrawLine(b->x+newleft, b->y, b->x+newright, b->y);
	    WinDrawLine(b->x+newleft, b->y+2, b->x+newright, b->y+2);
	}
	b->curleft = newleft;
	b->curright = newright;
    }
}

static void ScrollDraw(ScrollInfo *b)
{
    int l = b->curleft;
    int r = b->curright;
    /* Doing this will guarantee a redraw */
    b->curleft = 0;
    b->curright = b->barlength;
    if (b->hv == SCROLL_ORIENT_VERT) {
	WinDrawLine(b->x+1, b->y, b->x+1, b->y+b->barlength);
    } else {
	WinDrawLine(b->x, b->y+1, b->x+b->barlength, b->y+1);
    }
    ScrollUpdate(l, r, b, 1);
}

static void ScrollRepos(int newpos, ScrollInfo *b)
{
    int l = b->len > 0 ? (ULong)newpos * (ULong)b->barlength / (ULong)b->len
			: 0;
    int r = b->len > newpos + b->pagesize ?
		    (ULong)(newpos + b->pagesize) *
			(ULong)b->barlength / (ULong)b->len
		    : b->barlength;
    b->pos = newpos;
    ScrollUpdate(l, r, b, 0);
}

void ScrollPage(SWord xamt, SWord yamt)
{
    int newx = currXPosOffset;
    int newy = currYPosOffset;

    if (!hPage) return;

    /* Calculate the new x and y scroll positions, based on the requested
       relative scroll amounts, taking into account the total dimensions of
       the page. */
    if (yamt < 0) {
	if (currYPosOffset < -yamt) {
	    newy = 0;
	} else {
	    newy = currYPosOffset + yamt;
	}
    } else {
	if (vBar.len < 0) {
	    SWord width, height;
	    GetCurPageDims(&width, &height);
	    InitInfo(&vBar, height);
	    InitInfo(&hBar, width);
	}
	if (currYPosOffset + yamt + vBar.pagesize > vBar.len) {
	    if (vBar.pagesize < vBar.len) {
		newy = vBar.len - vBar.pagesize;
	    } else {
		newy = 0;
	    }
	} else {
	    newy = currYPosOffset + yamt;
	}
    }

    if (xamt < 0) {
	if (currXPosOffset < -xamt) {
	    newx = 0;
	} else {
	    newx = currXPosOffset + xamt;
	}
    } else {
	if (hBar.len < 0) {
	    SWord width, height;
	    GetCurPageDims(&width, &height);
	    InitInfo(&vBar, height);
	    InitInfo(&hBar, width);
	}
	if (currXPosOffset + xamt + hBar.pagesize > hBar.len) {
	    if (hBar.pagesize < hBar.len) {
		newx = hBar.len - hBar.pagesize;
	    } else {
		newx = 0;
	    }
	} else {
	    newx = currXPosOffset + xamt;
	}
    }

    /* Now scroll the page by blitting the appropriate rectangle, and then
       erasing and redrawing only the remaining gnomon.  This eliminates
       most of the "flicker" effect. */
    if (newy != currYPosOffset || newx != currXPosOffset) {
	RectangleType r1, r2, srcr;
	RectanglePtr rp1, rp2;
	WinHandle curwin;
	int destx, desty;
	int diffx, diffy;
	int adiffx, adiffy;

	diffx = newx - currXPosOffset;
	diffy = newy - currYPosOffset;
	adiffx = (diffx > 0) ? diffx : -diffx;
	adiffy = (diffy > 0) ? diffy : -diffy;
	if (adiffx >= MAIN_DISPLAY_WIDTH || adiffy >= MAIN_DISPLAY_HEIGHT) {
	    currXPosOffset = newx;
	    currYPosOffset = newy;
	    EraseMainWindow();
	    WinDrawProxyString(hPage, currXPosOffset, currYPosOffset);
	} else {
	    if (newy > currYPosOffset) {
		r1.topLeft.x = 0;
		r1.extent.x = MAIN_DISPLAY_WIDTH;
		srcr.extent.y = r1.topLeft.y = MAIN_DISPLAY_HEIGHT-adiffy;
		srcr.topLeft.y = r1.extent.y = adiffy;
		desty = 0;
		rp1 = &r1;
	    } else if (newy < currYPosOffset) {
		r1.topLeft.x = 0;
		r1.extent.x = MAIN_DISPLAY_WIDTH;
		srcr.topLeft.y = r1.topLeft.y = 0;
		desty = r1.extent.y = adiffy;
		srcr.extent.y = MAIN_DISPLAY_HEIGHT-adiffy;
		rp1 = &r1;
	    } else {
		srcr.topLeft.y = 0;
		srcr.extent.y = MAIN_DISPLAY_HEIGHT;
		desty = 0;
		rp1 = NULL;
	    }
	    if (newx > currXPosOffset) {
		srcr.extent.x = r2.topLeft.x = MAIN_DISPLAY_WIDTH-adiffx;
		srcr.topLeft.x = r2.extent.x = adiffx;
		destx = 0;
		r2.topLeft.y = 0;
		r2.extent.y = MAIN_DISPLAY_HEIGHT;
		rp2 = &r2;
	    } else if (newx < currXPosOffset) {
		srcr.topLeft.x = r2.topLeft.x = 0;
		destx = r2.extent.x = adiffx;
		srcr.extent.x = MAIN_DISPLAY_WIDTH-adiffx;
		r2.topLeft.y = 0;
		r2.extent.y = MAIN_DISPLAY_HEIGHT;
		rp2 = &r2;
	    } else {
		srcr.topLeft.x = 0;
		srcr.extent.x = MAIN_DISPLAY_WIDTH;
		destx = 0;
		rp2 = NULL;
	    }
	    currXPosOffset = newx;
	    currYPosOffset = newy;
	    curwin = WinGetActiveWindow();
	    WinCopy2BitRectangle(curwin, &srcr, destx, desty, scrCopy);
	    if (rp1) WinEraseRectangle(rp1, 0);
	    if (rp2) WinEraseRectangle(rp2, 0);
	    WinDrawProxyStringGnomon(hPage, currXPosOffset, currYPosOffset,
		rp1, rp2);
	}
	ScrollRepos(currYPosOffset, &vBar);
	ScrollRepos(currXPosOffset, &hBar);
    }
}

/* Reset the scroll bar */
void ScrollReset(void)
{
    SWord width, height;
    GetCurPageDims(&width, &height);
    InitInfo(&vBar, height);
    InitInfo(&hBar, width);
    ScrollDraw(&vBar);
    ScrollDraw(&hBar);
    currXPosOffset = 0;
    currYPosOffset = 0;
}

/* Refresh the scroll bar */
void ScrollRefresh(void)
{
    if (vBar.len < 0) {
	SWord width, height;
	GetCurPageDims(&width, &height);
	InitInfo(&vBar, height);
	InitInfo(&hBar, width);
    }
    ScrollDraw(&vBar);
    ScrollDraw(&hBar);
}

/* Handle an event */
Boolean ScrollHandleEvent(EventPtr e)
{
    Boolean handled = false;
    static int hdragging = 0;
    static int hdragloc = 0;
    static int vdragging = 0;
    static int vdragloc = 0;

    /* Is this interesting? */
    switch(e->eType) {
    case penDownEvent:
	vdragging = 0;
	hdragging = 0;
	if (e->screenX >= vBar.x && e->screenX <= vBar.x+2 &&
	    e->screenY >= vBar.y && e->screenY <= vBar.y+vBar.barlength) {
	    if (e->screenY < vBar.y + vBar.curleft) {
		ScrollPage(0, -vBar.pagesize);
	    } else if (e->screenY > vBar.y + vBar.curright) {
		ScrollPage(0, vBar.pagesize);
	    } else {
		/* Start a drag */
		vdragging = 1;
		vdragloc = e->screenY;
	    }
	    handled = true;
	} else if (e->screenX >= hBar.x && e->screenX <= hBar.x+hBar.barlength
	    && e->screenY >= hBar.y && e->screenY <= hBar.y+2) {
	    if (e->screenX < hBar.x + hBar.curleft) {
		ScrollPage(-hBar.pagesize, 0);
	    } else if (e->screenX > hBar.x + hBar.curright) {
		ScrollPage(hBar.pagesize, 0);
	    } else {
		/* Start a drag */
		hdragging = 1;
		hdragloc = e->screenX;
	    }
	    handled = true;
	}
	break;
    
    case penMoveEvent:
	if (vdragging) {
	    int sy = e->screenY;
	    int diff;
	    if (sy < vBar.y) sy = vBar.y;
	    if (sy > vBar.y + vBar.barlength) sy = vBar.y + vBar.barlength;
	    diff = sy - vdragloc;
	    if ((diff > 0 ? diff : -diff) > SCROLL_THRESH) {
		ScrollPage(0, (Long)diff * (Long)vBar.len /
			    (Long)vBar.barlength);
		vdragloc = sy;
	    }
	    handled = true;
	} else if (hdragging) {
	    int sx = e->screenX;
	    int diff;
	    if (sx < hBar.x) sx = hBar.x;
	    if (sx > hBar.x + hBar.barlength) sx = hBar.x + hBar.barlength;
	    diff = sx - hdragloc;
	    if ((diff > 0 ? diff : -diff) > SCROLL_THRESH) {
		ScrollPage((Long)diff * (Long)hBar.len /
			    (Long)hBar.barlength, 0);
		hdragloc = sx;
	    }
	    handled = true;
	}
	break;

    case penUpEvent:
	if (vdragging) {
	    int sy = e->screenY;
	    int diff;
	    if (sy < vBar.y) sy = vBar.y;
	    if (sy > vBar.y + vBar.barlength) sy = vBar.y + vBar.barlength;
	    diff = sy - vdragloc;
	    ScrollPage(0, (Long)diff * (Long)vBar.len / (Long)vBar.barlength);
	    vdragging = 0;
	    handled = true;
	} else if (hdragging) {
	    int sx = e->screenX;
	    int diff;
	    if (sx < hBar.x) sx = hBar.x;
	    if (sx > hBar.x + hBar.barlength) sx = hBar.x + hBar.barlength;
	    diff = sx - hdragloc;
	    ScrollPage((Long)diff * (Long)hBar.len / (Long)hBar.barlength, 0);
	    hdragging = 0;
	    handled = true;
	}
	break;
    
    default:
	break;
    }
    return handled;
}
