#pragma pack(2)

#include <Pilot.h>
#include "globals.h"
#include "prefs.h"
#include "callback.h"
#include "editmenu.h"
#include "utils.h"
#include "notify.h"

void ReadPrefs() {
  SWord size;
  void * prefs;
  Word prefSize = 20;
  Word proxySize = 64;

  prefs = MemPtrNew(prefSize);
  /* read proxy port and addr */
  size = PrefGetAppPreferences(APPTYPE, PROXY_PORT_PREF, prefs,
			       &prefSize, true);
  if (size != noPreferenceFound) {
    proxyPort = *((int *)prefs);
    size = PrefGetAppPreferences(APPTYPE, PROXY_ADDR_PREF, proxyServer,
				 &proxySize, true);
    if (size != noPreferenceFound) {
      /* NULL */
    }
  }
  /* read fallback proxy port and addr */
  size = PrefGetAppPreferences(APPTYPE, FALLBACK_PORT_PREF, prefs,
			       &prefSize, true);
  if (size != noPreferenceFound) {
    fallbackPort = *((int *)prefs);
    size = PrefGetAppPreferences(APPTYPE, FALLBACK_ADDR_PREF, fallbackServer,
				 &proxySize, true);
    if (size != noPreferenceFound) {
      /* NULL */
    } else {
      fallbackServer[0] = 0;
    }
  } else {
    fallbackPort = 0;
  }
  size = PrefGetAppPreferences(APPTYPE, HIST_PREF, prefs, &prefSize, true);
  if (size != noPreferenceFound) {    
    histSize = *((UInt *) prefs);
  }
  size = PrefGetAppPreferences(APPTYPE, CACHE_PREF, prefs, &prefSize, true);
  if (size != noPreferenceFound) {
    cacheSize = *((UInt *) prefs);
    cacheSize *= 1024;
  }
  size = PrefGetAppPreferences(APPTYPE, SCROLL_PREF, prefs, &prefSize, true);
  if (size != noPreferenceFound) { 
    scrollSize = *((UInt *) prefs);
  }
  size = PrefGetAppPreferences(APPTYPE, REMAP_PREF, prefs, &prefSize, true);
  if (size != noPreferenceFound) {    
    remap = *((Boolean *) prefs);
  }

  MemPtrFree(prefs);
}

void SetPrefs() {

  ULong t = cacheSize / (ULong) 1024;
  UInt tc = (UInt) t;

  PrefSetAppPreferences(APPTYPE, REMAP_PREF, 1, &remap, sizeof(Boolean), true);
  PrefSetAppPreferences(APPTYPE, PROXY_PORT_PREF, 1, &proxyPort,
			sizeof(int), true);
  PrefSetAppPreferences(APPTYPE, PROXY_ADDR_PREF, 1, &proxyServer,
			64, true);
  PrefSetAppPreferences(APPTYPE, FALLBACK_PORT_PREF, 1, &fallbackPort,
			sizeof(int), true);
  PrefSetAppPreferences(APPTYPE, FALLBACK_ADDR_PREF, 1, &fallbackServer,
			64, true);
  PrefSetAppPreferences(APPTYPE, HIST_PREF, 1, &histSize, sizeof(UInt), true);
  PrefSetAppPreferences(APPTYPE, CACHE_PREF, 1, &tc, sizeof(UInt), true);
  PrefSetAppPreferences(APPTYPE, SCROLL_PREF, 1, &scrollSize,
			sizeof(UInt), true);
}

Boolean PrefsFormHandleEvent (EventPtr e) {
  Boolean handled = false;
  FormPtr frm;
  FieldPtr fld, fld2;
  Char buf[10];
  ControlPtr ctl;

  CALLBACK_PROLOGUE

  switch (e->eType) {
  case frmOpenEvent:    
    frm = FrmGetActiveForm();
    fld = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsHistFld));    
    StrIToA(buf, histSize);
    SetFld(fld, buf, 4);
    fld = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsCacheFld));
    StrIToA(buf, (cacheSize/1024));
    SetFld(fld, buf, 4);
    fld = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsScrollFld));    
    StrIToA(buf, scrollSize);
    SetFld(fld, buf, 4);
    ctl = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsRemapBox));    
    CtlSetValue(ctl, remap);

    fld = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsProxyAddrFld));
    SetFld(fld, proxyServer, 64);

    fld = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsProxyPortFld));
    StrIToA(buf, proxyPort);
    SetFld(fld, buf, 5);

    if (fallbackServer[0]) {
      fld = FrmGetObjectPtr (frm,
			     FrmGetObjectIndex 
			     (frm, PrefsFallbackAddrFld));    
      SetFld(fld, fallbackServer, 64);
    }

    if (fallbackPort) {
      fld = FrmGetObjectPtr (frm, 
			     FrmGetObjectIndex 
			     (frm, PrefsFallbackPortFld));    
      StrIToA(buf, fallbackPort);
      SetFld(fld, buf, 5);
    }

    FrmDrawForm(frm);
    handled = true;
    break;
  
  case menuEvent:
    MenuEraseStatus(NULL);
    if (EditMenuHandleEvent(e->data.menu.itemID, FrmGetActiveForm())) {
	handled = true;
    } else {
	switch(e->data.menu.itemID) {
	}
    }
    handled = true;
    break;

  case ctlSelectEvent:
    switch(e->data.ctlEnter.controlID) {
    case PrefsSave:
      frm = FrmGetActiveForm();
      ctl = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsRemapBox));    
      remap = false;
       if (CtlGetValue(ctl)) remap = true;
       fld = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsScrollFld));
      scrollSize = StrAToI(FldGetTextPtr(fld));
      if (scrollSize > MAIN_DISPLAY_HEIGHT) {
	scrollSize = MAIN_DISPLAY_HEIGHT;
      }
      if (scrollSize < 1) {
	scrollSize = 1;
      }
      fld = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsHistFld));
      histSize = StrAToI(FldGetTextPtr(fld));
      if (histSize > 100) {
	histSize = 100;
      }
      if (histSize < 1) {
	histSize = 1;
      }
      fld = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsCacheFld));
      cacheSize = 1024 *  StrAToI(FldGetTextPtr(fld));
      fld = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsProxyPortFld));
      proxyPort = StrAToI(FldGetTextPtr(fld));
      fld2 = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, PrefsProxyAddrFld));
      StrCopy(proxyServer, FldGetTextPtr(fld2));

      fld = FrmGetObjectPtr (frm, FrmGetObjectIndex 
			     (frm, PrefsFallbackPortFld));
      if (FldGetTextPtr(fld)) {
	fallbackPort = StrAToI(FldGetTextPtr(fld));
      } else {
	fallbackPort = 0;
      }
      fld2 = FrmGetObjectPtr (frm, FrmGetObjectIndex 
			      (frm, PrefsFallbackAddrFld));
      if (FldGetTextPtr(fld2)) {
	StrCopy(fallbackServer, FldGetTextPtr(fld2));
      } else {
	fallbackServer[0] = 0;
      }
      SetPrefs();

      if (conID != -1) {
	CloseConnection(conID, true);
	conID = -1;
	conID = OpenConnection(proxyServer, proxyPort, storageDB);
	if ((conID == -1) && fallbackServer[0] && fallbackPort) {
	  conID = OpenConnection(fallbackServer, fallbackPort, storageDB);
	}
	if (conID != -1) {
	  RegisterNotify(conID);
	}
      }
  

      /* FALLTHROUGH */

    case PrefsCancel:
      if (firstOpen == 1) {
	  FrmGotoForm(SplashForm); 
      } else {
	  FrmGotoForm(MainForm); 
      }
      handled = true;
    break;
    }

  default:
    break;
  }

  CALLBACK_EPILOGUE

  return handled;
}
