#include <Pilot.h>
#include <sys/socket.h>
#include "globals.h"
#include "trans.h"
#include "send.h"
#include "cache.h"
#include "notify.h"
#include "metadata.h"
#include "pagedisplay.h"
#include "aggform.h"
#include "history.h"

/* the following function is an ugly hack; used to support stop loadings;
   should be implemented properly in next release */

void cancelSend(ConID conID) {
  if (conID != -1) {
      CloseConnection(conID, false);      
      StatusDraw(1);
      tcurHistory = -1;
  }
}

/* 
** returns true if it actually sends the url over the connection
** returns false if the request is ignored (becasue connection not opened)
*/
static Boolean SendURLNoCache(ConID *conID, VoidHand URL, Boolean nocache) {
  Chunk c;
  Word URLLength;
  Word err;
  Word but;

  /* if we have no conection yet, make one */
  if (*conID == -1) {
    FormPtr frm;

    frm = FrmInitForm(OpenConForm);
    but = FrmDoDialog(frm);
    if (but == OpenYes) {
      /* start network */      
      err = StartTrans();
      if (err != netErrAlreadyOpen) {
	ErrFatalDisplayIf(err, "Error opening transport layer");     
      }
      *conID = OpenConnection(proxyServer, proxyPort, storageDB);     
      if ((*conID == -1) && fallbackServer[0] && fallbackPort) {
	*conID = OpenConnection(fallbackServer, fallbackPort, storageDB);     
      }
      if (*conID == -1) {
	FrmDeleteForm(frm);
	StopTrans();
	frm = FrmInitForm(CantOpenConForm);
	FrmDoDialog(frm);
      } else {
	RegisterNotify(*conID);
      }    
    } 
    FrmDeleteForm(frm);
  }
  if (*conID != -1) {
    VoidHand metahand = NULL;
    MetaField metafields[2];
    int nummeta = 0;

    URLLength = MemHandleSize(URL);    
    c.conID = *conID;
    c.requestID = 0; /* set to zero to let transport layer deal with it */
    if (nocache) {
	metafields[nummeta].type = MetaNoCache;
	metafields[nummeta].len = 0;
	metafields[nummeta].data = NULL;
	++nummeta;
    }
    if (NeedAggList()) {
	metafields[nummeta].type = MetaAggList;
	metafields[nummeta].len = 0;
	metafields[nummeta].data = NULL;
	++nummeta;
    }
    c.metaHand = metahand = MakeMetaData(metafields, nummeta, &(c.metaSize));
    c.dataSize = URLLength;
    c.dataHand = URL;
    SendChunk(&c);
    if (metahand) {
	MemHandleFree(metahand);
    }
    currRequestID = c.requestID;
    return true;
  }
  return false;
}

/* 
** returns true (and sets hPage) if the URL was in the cache
** returns false otherwise
*/
static Boolean LookInCacheForURL(VoidHand URL) {
  Word URLLength;
  char * url;
  VoidHand data;

  URLLength = MemHandleSize(URL);
  url = MemHandleLock(URL);
  data = InCache(url, URLLength, offlineInfoDB, offlineDB);
  if (!data) {
    data = InCache(url, URLLength, cacheInfoDB, cacheDB);
  }
  MemHandleUnlock(URL);
  if (data) {
    hPage = data;
    UpdateHiLiPiece(-2, NULL, 0, 0);
    ScrollReset();
    return true;
  }
  return false;
}


/* 
**  returns 2 if the data was found in cache..
**  1 if the data was sent off to the net...
**  0 if the request was never made
**  sets hapge and all those icky tcurURL and curURL fields
*/

/* ### HEY DUMBASS, FIGURE OUT WHAT ALL THIS tcurURL transURL stuff IS ### */
/* ### RELOAD!  RELOAD!  ACK!  ACK!  ACK! ### */
int SendURL (ConID *conID, VoidHand URL, Boolean addToHist, 
	     Boolean checkCache, Boolean nocache, Boolean changeURL) {

  ULong URLLength;

  URLLength = MemHandleSize(URL);

  if (*conID > -1) {
    cancelSend(*conID);
  }
  if (checkCache && LookInCacheForURL(URL)) {
    if (changeURL) {
      if (curURL) {
	MemHandleFree(curURL);
      }
      curURL = URL;
      curURLLength = URLLength;
    }
    if (addToHist) {
      PutInHistory(URL, URLLength);
    }
    EraseMainWindow();
    if (hPage) {
      WinDrawProxyString(hPage, currXPosOffset, currYPosOffset);
    }   
    return 2;
  } else {
    if (SendURLNoCache(conID, URL, nocache)) {
      if (changeURL) {
	if (transURL) {
	  MemHandleFree(transURL);
	}
	transURL = URL;
	transURLLength = URLLength;  
      }
      if (addToHist) {
	tcurHistory = curHistory;	
      } else {
	tcurHistory = -1;	
      }
      return 1;
    } else {
      return 0;
    }    
  }	 
}

