#include <Pilot.h>
#include "2bit.h"
#include "start.h"
#include "trans.h"
#include "globals.h"
#include "prefs.h"
#include "notify.h"

/* Set up the display */
void displayInit() {
  RectangleType dispBounds;
  
  /* Set up the 1-bit window */
  dispBounds.topLeft.x = 0; 
  dispBounds.topLeft.y = 0;
  dispBounds.extent.x = 160;
  dispBounds.extent.y = 160;
  disp = WinCreateWindow(&dispBounds, noFrame, 0, 1, &dispError);
  WinInitializeWindow(disp);
  WinSetDrawWindow(disp);

  /* Go 2-bit */
#ifndef NOTWOBIT
  enable_2bit();
#endif
}  

/* Get preferences, open (or create) app database */
Word StartApplication(void)
{
  Word err;

  SetGlobals();

  historyDB = DmOpenDatabaseByTypeCreator('Hist', APPTYPE, dmModeReadWrite);  
  if (historyDB == NULL) {
    err = DmCreateDatabase(0, "WingmanHistory", APPTYPE, 'Hist', 0);
    ErrFatalDisplayIf(err, "Error creating history database");
    historyDB = DmOpenDatabaseByTypeCreator('Hist', APPTYPE, dmModeReadWrite);
    ErrFatalDisplayIf(!historyDB, "Error opening history database");
  }

  cacheDB = DmOpenDatabaseByTypeCreator('Cash', APPTYPE, dmModeReadWrite);  
  if (cacheDB == NULL) {
    err = DmCreateDatabase(0, "WingmanCache", APPTYPE, 'Cash', 0);
    ErrFatalDisplayIf(err, "Error creating cache database");
    cacheDB = DmOpenDatabaseByTypeCreator('Cash', APPTYPE, dmModeReadWrite);
    ErrFatalDisplayIf(!cacheDB, "Error opening cache database");
  }

  cacheInfoDB = DmOpenDatabaseByTypeCreator('cInf', APPTYPE, dmModeReadWrite);
  if (cacheInfoDB == NULL) {
    err = DmCreateDatabase(0, "WingmanCacheInfo", APPTYPE, 'cInf', 0);
    ErrFatalDisplayIf(err, "Error creating cache info database");
    cacheInfoDB = DmOpenDatabaseByTypeCreator('cInf', 
					      APPTYPE, dmModeReadWrite);
    ErrFatalDisplayIf(!cacheInfoDB, "Error opening cache info database");
  }

  offlineDB = DmOpenDatabaseByTypeCreator('Offl', APPTYPE, dmModeReadWrite);  
  if (offlineDB == NULL) {
    err = DmCreateDatabase(0, "WingmanOfflineCache", APPTYPE, 'Offl', 0);
    ErrFatalDisplayIf(err, "Error creating offline cache database");
    offlineDB = DmOpenDatabaseByTypeCreator('Offl', APPTYPE, dmModeReadWrite);
    ErrFatalDisplayIf(!offlineDB, "Error opening offline cache database");
  }

  offlineInfoDB = DmOpenDatabaseByTypeCreator('oInf', APPTYPE, 
					      dmModeReadWrite);
  if (offlineInfoDB == NULL) {
    err = DmCreateDatabase(0, "WingmanOfflineCacheInfo", APPTYPE, 'oInf', 0);
    ErrFatalDisplayIf(err, "Error creating offline cache info database");
    offlineInfoDB = DmOpenDatabaseByTypeCreator('oInf', 
						APPTYPE, dmModeReadWrite);
    ErrFatalDisplayIf(!offlineInfoDB, 
		      "Error opening offline cache info database");
  }


  hotlistDB = DmOpenDatabaseByTypeCreator('Hotl', APPTYPE, dmModeReadWrite);  
  if (hotlistDB == NULL) {
    err = DmCreateDatabase(0, "WingmanHotList", APPTYPE, 'Hotl', 0);
    ErrFatalDisplayIf(err, "Error creating hot list database");
    hotlistDB = DmOpenDatabaseByTypeCreator('Hotl', APPTYPE, dmModeReadWrite);
    ErrFatalDisplayIf(!hotlistDB, "Error opening hot list database");
  }
  
  storageDB = DmOpenDatabaseByTypeCreator('STOR', APPTYPE, dmModeReadWrite);  
  if (storageDB == NULL) {
    err = DmCreateDatabase(0, "WingmanTransStorage", APPTYPE, 'STOR', 0);
    ErrFatalDisplayIf(err, "Error creating storage database");
    storageDB = DmOpenDatabaseByTypeCreator('STOR', APPTYPE, dmModeReadWrite);
    ErrFatalDisplayIf(!storageDB, "Error opening storage database");
  } else {
    /* storage database exists!  clean it out */
    ULong numRecs;
    LocalID lid;
    UInt cardno;
    ULong i;

    DmOpenDatabaseInfo(storageDB, &lid, NULL, NULL, &cardno, NULL);
    DmDatabaseSize(cardno, lid, &numRecs, NULL, NULL);
    
    for (i = 0; i < numRecs; i++) {
      DmRemoveRecord(storageDB, 0);
    }
  }

  ReadPrefs();

  /* start display */
  displayInit();

  StatusInit();

  FrmGotoForm(SplashForm);
  return 0;
}
