#include <Pilot.h>

#include <sys/socket.h>
#include "globals.h"
#include "trans.h"


/* header is 5 32 bit unsigned words long */
/* version                                */
/* request id                             */
/* metaSize                               */
/* dataSize                               */
/* compression type                       */
#define HEADERSIZE 20
/* version of transport/program */
#define VERSIONLOC 0
/* request ID */
#define RIDLOC 4
/* meta data size */
#define METALOC 8
/* regular data size */
#define SIZELOC 12
/* compression type */
#define COMPLOC 16

#define ConID2Socket(conID) ((Int) conRecs[conID].sock)


/*ReadCallbackFuction*/
typedef void (*Rcf)(Int, unsigned char *, int, void *);

typedef struct {
  Int sock;

  Boolean inUse;

  struct sockaddr_in saddr;
  DmOpenRef db;
  
  RequestID nextRequest;
  ChunkID nextChunk;

  /* and now the info needed for reading data */
  /* getting kinda large here                 */
  VoidHand dataBufHand;
  VoidHand metaBufHand;
  int buffed;
  unsigned char headBuf[HEADERSIZE];
  DataSize dataSize;
  DataSize metaSize;
  RequestID requestID;
  DataSize left;
  DataSize seen;
  Boolean readingMeta;
  DataCallback callback;
  DataSize * total;
  DataSize * sofar;
  void * ptr;
  DWord version;

  /* and now some stuff for database evilness... THIS IS GETTING HUGE */
  ULong metaID;
  ULong dataID;

} ConRec;

static ConRec conRecs[MAX_CONNECTIONS];

Err errno;

/* returns 0 if all ok, -1 if it can't allocate the memory needed */
static int MakeDataHandle(DataSize *ds, VoidHand *dh, 
			    ULong *id, DmOpenRef db) {
  Err e;
  UInt at = 0;
  if (*ds > 0xFFFF) {
    /* 
     *  we want to read data that is larger than a short, 
     *  right now this is illegal
     */
    *dh = NULL;
    *ds = 0;
    return -1;
  } else if (*ds) {	 
    /* if there is data, make the handle for it */
    *dh = DmNewRecord(db, &at, *ds);
    if (!(*dh)) {      
      /* not enough space */
      return -1;
    }
    e = DmRecordInfo(db, at, NULL, id, NULL);
    ErrFatalDisplayIf(e, "Trans: Error finding data record");
  } else  {
    /* no data to store, therefore don't allocate a handle to store it in*/
    *dh = NULL;
  }
  return 0;
}

Err DestroyChunk(ChunkHand ch) {
  UInt index;
  ChunkPtr c;
  Err err;
  
  c = (ChunkPtr) MemHandleLock(ch);

  if ((c->conID < FIRST_CONNECTION) || (c->conID >= MAX_CONNECTIONS)) {
    return 1;
  } 

  if (conRecs[c->conID].db) {
    DmOpenRef db = conRecs[c->conID].db;

    if (c->dataSize) {
      err = DmFindRecordByID(db, c->dataID, &index);
      if (err) return err;
      err = DmRemoveRecord(db, index);
      if (err) return err;      
    }
    if (c->metaSize) {
      err = DmFindRecordByID(db, c->metaID, &index);
      if (err) return err;      
      err = DmRemoveRecord(db, index);
      if (err) return err;      
    }
  } else {
    if (c->dataSize) {
      err = MemHandleFree(c->dataHand);
      if (err) return err;      
    }
    if (c->metaSize) {
      err = MemHandleFree(c->metaHand);
      if (err) return err;      
    }
  }
  MemHandleUnlock(ch);
  err = MemHandleFree(ch);
  if (err) return err;

  return 0;
}

Int ReOpenConnection(ConID conID) {
  Err err;
  Byte allup;
  Word iferr;

  NetLibConnectionRefresh(AppNetRefnum, true, &allup, &iferr);  
  if (!allup) return -1;
  
  close(ConID2Socket(conID));
  
  conRecs[conID].sock = socket(PF_INET, SOCK_STREAM, 0);
  err = connect(conRecs[conID].sock, 
		(struct sockaddr *)&conRecs[conID].saddr, 
		sizeof(conRecs[conID].saddr));
  if (err) {
    conRecs[conID].sock = -1;
    return -1;
  }     
  return conRecs[conID].sock;
}

ConID Socket2ConID (Int s) {
  ConID id;
  for (id = 0; id < MAX_CONNECTIONS; id++) {
    if ((conRecs[id].sock == s) && (conRecs[id].inUse)) {
      return id;
    }
  }
  return -1;
}

Err StartTrans() {
  Word err = 0;
  UInt foo;
 
  AppNetTimeout = 1000;

  /* all connections are unused */
  for (foo = 0; foo < MAX_CONNECTIONS; foo++) {
    conRecs[foo].inUse = 0;
    conRecs[foo].left = 0;   
    conRecs[foo].buffed = 0;
    conRecs[foo].seen = 0;
    conRecs[foo].callback = NULL;
  }

  /* open the net library */
  err = SysLibFind("Net.lib", &AppNetRefnum);
  if (err) {
    return err;
  }
  err = NetLibOpen(AppNetRefnum, &foo);  
  return err;
}

Err StopTrans() {
  Err err = 0;

  if (AppNetRefnum) {
    err = NetLibClose(AppNetRefnum, 0);
  }
  return err;
}

static Int get_data(Int s, Rcf rc, void *rcp) {
  int read;
  char *buf;

  /* allocate us a buffer, but don't put it on the stack */
  buf = (char *) MemPtrNew(1024);

  read = read(s, buf, 1024);
  if (read <= 0) {
    conRecs[Socket2ConID(s)].sock = -1;
    MemPtrFree(buf);  
    return 1;
  } else {
    rc(s, buf, read, rcp);
    MemPtrFree(buf);
    return 0;
  }
}

static Boolean SendData(Int * s, DataSize size, unsigned char * data) {
  while (send(*s, data, size, 0) <= 0) {    

    *s = ReOpenConnection(Socket2ConID(*s));
    if (*s == -1) {
      return 1;
    }
  }
  return 0;
}

Err SendChunk(ChunkPtr chunk) {
  Err err;
  Int s;
  CompType compType = TRANS_NO_COMP;
  ConRec *cr;

  if ((chunk->conID >= MAX_CONNECTIONS) || (chunk->conID < FIRST_CONNECTION)) {
    return 1;
  }

  if (!chunk->requestID) {
    chunk->requestID = conRecs[chunk->conID].nextRequest;
    conRecs[chunk->conID].nextRequest++;
    conRecs[chunk->conID].nextChunk = 1;
  }

  if (!chunk->chunkID) {
    chunk->chunkID = conRecs[chunk->conID].nextChunk;
    conRecs[chunk->conID].nextChunk++;
  }

  s = ConID2Socket(chunk->conID);

  /* Notify that we're sending it */
  cr = &conRecs[conID];
  if (cr->callback) {
    if (cr->total) {
      *(cr->total) = 1;
    }
    if (cr->sofar) {
      *(cr->sofar) = 0;
    }
    cr->callback(cr->total, cr->sofar, cr->ptr);
  }

  err = SendData(&s, sizeof(DWord), (unsigned char *) 
		 (&(version)));
  if (err) {
    /*ErrFatalDisplayIf(err, "Error sending data");*/
    return err;
  }

  err = SendData(&s, sizeof(RequestID), (unsigned char *) 
		 (&(chunk->requestID)));
  if (err) {
    /*ErrFatalDisplayIf(err, "Error sending data");*/
    return err;
  }
  err = SendData(&s, sizeof(DataSize), (unsigned char *) (&(chunk->metaSize)));
  if (err) {
    /*ErrFatalDisplayIf(err, "Error sending data");*/
    return err;
  }
  err = SendData(&s, sizeof(DataSize), (unsigned char *) (&(chunk->dataSize)));
  if (err) {
    /*ErrFatalDisplayIf(err, "Error sending data");*/
    return err;
  }
  err = SendData(&s, sizeof(CompType), (unsigned char *) (&compType));
  if (err) {
    /*ErrFatalDisplayIf(err, "Error sending data");*/
    return err;
  }
  if (chunk->metaSize) {
    err = SendData(&s, chunk->metaSize, (unsigned char *) 
		   MemHandleLock(chunk->metaHand));
    MemHandleUnlock(chunk->metaHand);
  
    if (err) {
      /*ErrFatalDisplayIf(err, "Error sending data");*/
      return err;
    }
  }
  if (chunk->dataSize) {
    err = SendData(&s, chunk->dataSize, (unsigned char *) 
		   MemHandleLock(chunk->dataHand));
    MemHandleUnlock(chunk->dataHand);
    if (err) {
      /*ErrFatalDisplayIf(err, "Error sending data");*/
      return err;
    }
  }
  return 0;
}

static void NewChunkCallback(ConID conID, RequestID requestID, 
			     DataSize metaSize, DataSize dataSize, 
			     VoidHand metaHand, VoidHand dataHand, 
			     ULong metaID, ULong dataID,
			     VoidPtr ncp) {
  ChunkHand ch = MemHandleNew(sizeof(Chunk));
  transEvent te;

  ChunkPtr c = (ChunkPtr) MemHandleLock(ch);

  conRecs[conID].nextChunk++;

  c->conID = conID;
  c->requestID = requestID;
  c->chunkID = conRecs[conID].nextChunk;

  c->metaHand = metaHand;
  c->metaSize = metaSize;
  
  c->dataHand = dataHand;
  c->dataSize = dataSize;

  c->dataID = dataID;
  c->metaID = metaID;

  MemHandleUnlock(ch);
  
  te.eType = transNewChunk;
  te.data.newChunk.chunkHand = ch;
  EvtAddEventToQueue((EventPtr) &te);
}

static void ReadCallback(Int s, unsigned char *buf, int amt, void *rcp) {
  /* as in read.  is just size-left but this makes me happier...OK? */
  ConID conID = Socket2ConID(s);

  char *dataPtr;
  int *signal = rcp;
  int i;
  ConRec *cr = &conRecs[conID];

  /* set break to debug */

  while(amt) {
    /* read the entire header before doing anything stupid */
    if (cr->left == 0) {
      if ((amt+cr->buffed) < HEADERSIZE) {
	for (i = 0; i < amt; i++) {
	  cr->headBuf[i+cr->buffed] = buf[i];
	}
	cr->buffed += amt;
	amt = 0;
      } else {
	for (i = 0; cr->buffed+i < HEADERSIZE; i++) {      
	  cr->headBuf[i+cr->buffed] = buf[i];
	}
	/* WE HAVE AN ENTIRE HEADER!  YAY!                              */
	/* warning this may crash the machine because of word alignment */
	/* but it doesn't seem to now... may later?                     */
	cr->version = htonl(*((long*) (cr->headBuf+VERSIONLOC)));
	ErrFatalDisplayIf(
	    ((cr->version & 0xffff0000) > (version & 0xffff0000)), 
	    "Protocol version mismatch");
	cr->requestID = htonl(*((long*) (cr->headBuf+RIDLOC)));
	cr->dataSize = htonl(*((long*) (cr->headBuf+SIZELOC)));
	cr->metaSize = htonl(*((long*) (cr->headBuf+METALOC)));

	if (cr->metaSize) {
	  cr->left = cr->metaSize;
	  cr->readingMeta = 1;
	} else if (cr->dataSize) {
	  cr->left = cr->dataSize;
	  cr->readingMeta = 0;
	} else {
	  /* all we are is header! */
	  cr->left = 0;
	  if (signal) *signal = 1;
	  NewChunkCallback(Socket2ConID(s), cr->requestID, 0, 
			   0, NULL, NULL, 0, 0,
			   NULL);
	  break;
	}

	cr->seen = 0;
	if (cr->db) {
	  if (MakeDataHandle(&(cr->dataSize), &(cr->dataBufHand),
			     &(cr->dataID),  cr->db)) {	    
	    /* ERROR MAKING SPACE FOR DATA... CLOSE CONNECTION */
	    if (cr->callback) {
	      if (cr->total) {
		*cr->total = 0;
	      }
	      if (cr->sofar) {
		*cr->sofar = 0;
	      }
	      cr->callback(cr->total, cr->sofar, cr->ptr);
	    }
	    CloseConnection(conID, false);
	    FrmAlert(OutOfMemoryAlert);
	    return;
	  } else if (MakeDataHandle(&(cr->metaSize), &(cr->metaBufHand), 
				    &(cr->metaID), cr->db)) {
	    /* ERROR MAKING SPACE FOR DATA... CLOSE CONNECTION */
	    if (cr->callback) {
	      if (cr->total) {
		*cr->total = 0;
	      }
	      if (cr->sofar) {
		*cr->sofar = 0;
	      }
	      cr->callback(cr->total, cr->sofar, cr->ptr);
	    }
	    CloseConnection(conID, false);
	    FrmAlert(OutOfMemoryAlert);
	  }
	} else {
	  if (cr->dataSize) {
	    cr->dataBufHand = MemHandleNew(cr->dataSize);
	    if (!(cr->dataBufHand)) {
	      CloseConnection(conID, false);
	      FrmAlert(OutOfMemoryAlert);	      
	      return;
	    }
	  }       
	  if (cr->metaSize) {
	    cr->metaBufHand = MemHandleNew(cr->metaSize);
	    if (!(cr->metaBufHand)) {
	      CloseConnection(conID, false);
	      FrmAlert(OutOfMemoryAlert);	      
	      return;
	    }
	  }
	}
	buf += i;
	amt -= i;
	cr->buffed=0;
      }
    } else {
      if (cr->readingMeta) {
	if (amt >= cr->left) {
	  dataPtr = MemHandleLock(cr->metaBufHand);
	  if (cr->db) {
	    DmWrite(dataPtr, cr->seen, buf, cr->left);
	  } else {
	    MemMove(dataPtr+cr->seen, buf, cr->left);
	  }
	  MemHandleUnlock(cr->metaBufHand);
	  cr->seen = 0;
	  amt -= cr->left;
	  buf += cr->left;
	  cr->left = cr->dataSize;
	  cr->readingMeta = 0;
	} else {
	  dataPtr = MemHandleLock(cr->metaBufHand);
	  if (cr->db) {
	    DmWrite(dataPtr, cr->seen, buf, amt);
	  } else {
	    MemMove(dataPtr+cr->seen, buf, amt);
	  }
	  MemHandleUnlock(cr->metaBufHand);
	  cr->seen += amt;
	  cr->left -= amt;
	  amt = 0;
	}
      } else {
	/* reading the actual data now */
	if (amt >= cr->left) {
	  dataPtr = MemHandleLock(cr->dataBufHand);
	  if (cr->db) {
	    DmWrite(dataPtr, cr->seen, buf, cr->left);
	  } else {
	    MemMove(dataPtr+cr->seen, buf, cr->left);
	  }
	  MemHandleUnlock(cr->dataBufHand);
	  NewChunkCallback(Socket2ConID(s), cr->requestID, cr->metaSize, 
			   cr->dataSize, cr->metaBufHand, cr->dataBufHand,
			   cr->metaID, cr->dataID, NULL);
	  /* Signal that we created an event */
	  cr->dataBufHand = NULL;
	  cr->metaBufHand = NULL;
	  if (signal) *signal = 1;
	  cr->seen += cr->left;
	  amt -= cr->left;
  buf += cr->left;
	  cr->left = 0;
	} else {
	  dataPtr = MemHandleLock(cr->dataBufHand);
	  if (cr->db) {
	    DmWrite(dataPtr, cr->seen, buf, amt);
	  } else {
	    MemMove(dataPtr+cr->seen, buf, amt);
	  }
	  MemHandleUnlock(cr->dataBufHand);
	  cr->seen += amt;
	  cr->left -= amt;
	  amt = 0;
	}
      }
      if (cr->callback) {
	if (cr->total) {
	  *(cr->total) = cr->metaSize + cr->dataSize;
	}
	if (cr->sofar) {
	  *(cr->sofar) = cr->readingMeta ? 
	    cr->seen : cr->metaSize + cr->seen;
	}
	cr->callback(cr->total, cr->sofar, 
			    cr->ptr);
      } else {
      }
    }
  }
}

ConID OpenConnection(char * host, unsigned short port, DmOpenRef db) { 
    struct hostent *h;
    ULong **addrp;
    Word err;
    Word iferr;
    Byte allup;
    ConID conID;
    
    NetLibConnectionRefresh(AppNetRefnum, true, &allup, &iferr);
    if (!allup) return -1;

    for (conID = 0; conID < MAX_CONNECTIONS; conID++) {
      if (!conRecs[conID].inUse) {
	break;
      }
    }
   if (conRecs[conID].inUse) {      
      return -1;      
    }

    conRecs[conID].db = db;

    h = gethostbyname(host);
    if (!h) {
      /*ErrFatalDisplayIf(1, host);*/
      return -1;
    }

    conRecs[conID].saddr.sin_family = AF_INET;
    conRecs[conID].saddr.sin_port = htons(port);
    
    for(addrp=(ULong **)h->h_addr_list;*addrp;++addrp) {
      ULong a = htonl(**addrp);
      if (a != INADDR_ANY) {
	conRecs[conID].saddr.sin_addr.s_addr = a;
	break;
      }
    }

    conRecs[conID].sock = socket(PF_INET, SOCK_STREAM, 0);
    if (conRecs[conID].sock < 0) {      
      /*ErrFatalDisplayIf(1, "Cannot create socket");*/
      return -1;
    }

    err = connect(conRecs[conID].sock, 
		  (struct sockaddr *)&conRecs[conID].saddr, 
		  sizeof(conRecs[conID].saddr));
    if (err) {
      /*ErrFatalDisplayIf(1, "Could not connect");*/
	return -1;
    }

    conRecs[conID].inUse = 1;
    conRecs[conID].nextRequest = 1;

    return conID;
}

void RegisterNotifyCallback(ConID conID, DataCallback cb, 
		      DataSize * total, DataSize * sofar, void * ptr) {
  if ((conID < MAX_CONNECTIONS) && (conID >= FIRST_CONNECTION)) {
    conRecs[conID].callback = cb;
    conRecs[conID].total = total;
    conRecs[conID].sofar = sofar;
    conRecs[conID].ptr = ptr;
  }
}

/* if final is true CloseConneciton will destroy the mapping between
   the conid and the server.  Otherwise it will just close the socket */
void CloseConnection(ConID conID, Boolean final) {
  ConRec *c;

  if ((conID >= MAX_CONNECTIONS) || (conID < FIRST_CONNECTION)) {
    return;
  }

  c = &conRecs[conID];

  close(ConID2Socket(conID));

  if (final) {
    c->inUse = 0;
    c->callback = NULL;
  }

  c->left = 0;   
  c->buffed = 0;
  c->seen = 0;
  if (c->db) {
    DmOpenRef db = c->db; 
    UInt index;
    if (c->dataBufHand) {
      DmFindRecordByID(db, c->dataID, &index);
      DmRemoveRecord(db, index);
      c->dataBufHand = NULL;
    }
    if (c->metaBufHand) {
      DmFindRecordByID(db, c->metaID, &index);
      DmRemoveRecord(db, index);
      c->metaBufHand = NULL;
    }
  } else {
    if (c->dataBufHand) {
      MemHandleFree(c->dataBufHand);
      c->dataBufHand = NULL;
    }
    if (c->metaBufHand) {
      MemHandleFree(c->metaBufHand);
      c->metaBufHand = NULL;
    }
  }
}

void CheckNetwork() {
  Int maxSock = 0;
  fd_set reads;
  UInt ok;
  int read;
  int gotone = 0;
  int i;

  do {
    /* Check if there's data available */
    read = 0;
    FD_ZERO(&reads);
    FD_SET(STDIN_FILENO,&reads);
    for (i = FIRST_CONNECTION; i < MAX_CONNECTIONS; i++) {
      if (conRecs[i].inUse && (conRecs[i].sock > 0)) {
	FD_SET(conRecs[i].sock,&reads);
	maxSock = maxSock > conRecs[i].sock ? maxSock : conRecs[i].sock;
      }
    }
    ok = NetLibSelect(AppNetRefnum, maxSock+1, &reads, NULL, NULL, 
		      gotone ? 0 : evtWaitForever, &errno);
    if (ok > 0) {
      for (i = FIRST_CONNECTION; i < MAX_CONNECTIONS; i++) {
	if (conRecs[i].inUse && FD_ISSET(conRecs[i].sock, &reads)) {
	  read = 1;
	  get_data(conRecs[i].sock, ReadCallback, &gotone);
	}
      }
    }
  } while (read);
}

